from collections import defaultdict

from django import template
from django.contrib.auth import get_permission_codename
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.utils.text import camel_case_to_spaces
from django.utils.translation import gettext_noop

from wagtail.admin.models import Admin
from wagtail.users.permission_order import get_content_type_order_lookup

register = template.Library()


def normalize_permission_label(permission: Permission):
    """
    Strip model name from the end of the label, e.g. "Can deliver pizza" for a
    Pizza model becomes "Can deliver". For permissions in the model's
    Meta.default_permissions with default labels, also replace underscores
    with spaces.

    This is used to display custom model permissions in the admin.

    See https://github.com/wagtail/wagtail/issues/10982.
    """
    label = permission.name
    content_type = permission.content_type
    model = content_type.model_class()
    verbose_name = default_verbose_name = content_type.name

    if model:
        default_verbose_name = camel_case_to_spaces(model._meta.object_name)

        # If it's in default_permissions and the label matches Django's default
        # label, remove the model name from the end of the label. Also replace
        # underscores with spaces, as Django uses the action internal name as-is
        # for the permission label, which means it tends to be in snake_case.
        for action in model._meta.default_permissions:
            default_codename = get_permission_codename(action, model._meta)
            is_default = permission.codename == default_codename
            if is_default and permission.name.startswith(f"Can {action}"):
                return f"Can {action.replace('_', ' ')}"

    # For all other cases (including custom permissions), try to remove the
    # verbose name from the end of the label. This only works if the label
    # matches the current verbose name or Django's default verbose name.
    for name in (default_verbose_name, verbose_name):
        if label.lower().endswith(name.lower()):
            return label[: -len(name)].strip()

    return label


# normalize_permission_label will return "Can view" for Django's standard "Can view X" permission.
# formatted_permissions.html passes these labels through {% trans %} - since this is a variable
# within the template it will not be picked up by makemessages, so we define a translation here
# instead.

VIEW_PERMISSION_LABEL = gettext_noop("Can view")


@register.inclusion_tag("wagtailusers/groups/includes/formatted_permissions.html")
def format_permissions(permission_bound_field):
    """
    Given a bound field with a queryset of Permission objects - which must be using
    the CheckboxSelectMultiple widget - construct a list of dictionaries for 'objects':

    'objects': [
        {
            'object': name_of_some_content_object,
            'add': checkbox,
            'change': checkbox,
            'delete': checkbox,
            'custom': list_of_checkboxes_for_custom_permissions
        },
    ]

    and a list of other permissions:

    'others': [
        (any_non_add_change_delete_permission, checkbox),
    ]

    (where 'checkbox' is an object with a tag() method that renders the checkbox as HTML;
    this is a BoundWidget on Django >=1.11)

    - and returns a table template formatted with this list.

    """
    permissions = permission_bound_field.field._queryset
    # get a distinct and ordered list of the content types that these permissions relate to.
    # relies on Permission model default ordering, dict.fromkeys() retaining that order
    # from the queryset, and the stability of sorted().
    content_type_order = get_content_type_order_lookup()
    content_type_ids = sorted(
        dict.fromkeys(permissions.values_list("content_type_id", flat=True)),
        key=lambda ct: content_type_order.get(ct, float("inf")),
    )

    # iterate over permission_bound_field to build a lookup of individual renderable
    # checkbox objects
    # checkbox.data['value'] gives a ModelChoiceIteratorValue
    checkboxes_by_id = {
        int(checkbox.data["value"].value): checkbox
        for checkbox in permission_bound_field
    }

    # Permissions that are known by Wagtail, to be shown under their own columns.
    # Other permissions will be shown under the "custom permissions" column.
    main_permission_names = ["add", "change", "delete"]

    # Only show the columns for these permissions if any of the model has them.
    extra_perms_exist = {
        "custom": False,
    }
    # Batch the permission query for all content types, then group by content type
    # (instead of querying permissions for each content type separately)
    content_perms_by_ct_id = defaultdict(list)
    permissions = permissions.filter(content_type_id__in=content_type_ids)
    for permission in permissions:
        content_perms_by_ct_id[permission.content_type_id].append(permission)

    # Permissions that use Wagtail's Admin content type, to be displayed
    # under the "Other permissions" section alongside the
    # "Can access Wagtail admin" permission.
    admin_content_type = ContentType.objects.get_for_model(Admin)
    admin_permissions = content_perms_by_ct_id.pop(admin_content_type.id, [])
    other_perms = [(perm, checkboxes_by_id[perm.id]) for perm in admin_permissions]

    # We're done with the admin content type, so remove it from the list of content types
    # but make sure the sorted order is preserved.
    content_type_ids = [
        ct_id for ct_id in content_type_ids if ct_id != admin_content_type.pk
    ]

    # Permissions for all other content types, to be displayed under the
    # "Object permissions" section.
    object_perms = []

    # Iterate using the sorted content_type_ids
    for ct_id in content_type_ids:
        content_perms = content_perms_by_ct_id[ct_id]
        content_perms_dict = {}
        custom_perms = []

        for perm in content_perms:
            content_perms_dict["object"] = perm.content_type.name
            checkbox = checkboxes_by_id[perm.id]
            attrs = {"data-action": "w-bulk#toggle", "data-w-bulk-target": "item"}
            # identify the main categories of permission, and assign to
            # the relevant dict key, else bung in the 'custom_perms' list
            permission_action = perm.codename.split("_")[0]
            is_known = (
                permission_action in main_permission_names
                and perm.codename == f"{permission_action}_{perm.content_type.model}"
            )

            if is_known:
                if permission_action in extra_perms_exist:
                    extra_perms_exist[permission_action] = True
                checkbox.data["attrs"].update(attrs)
                checkbox.data["attrs"]["data-w-bulk-group-param"] = permission_action
                content_perms_dict[permission_action] = {
                    "perm": perm,
                    "checkbox": checkbox,
                }
            else:
                extra_perms_exist["custom"] = True
                attrs["data-w-bulk-group-param"] = "custom"
                perm_name = normalize_permission_label(perm)
                custom_perms.append(
                    {
                        "attrs": attrs,
                        "perm": perm,
                        "name": perm_name,
                        "selected": checkbox.data["selected"],
                    }
                )

        content_perms_dict["custom"] = custom_perms
        object_perms.append(content_perms_dict)
    return {
        "object_perms": object_perms,
        "other_perms": other_perms,
        "extra_perms_exist": extra_perms_exist,
    }



import django_filters
from django.contrib.auth import (
    get_user_model,
    update_session_auth_hash,
)
from django.contrib.auth.models import Group
from django.core.exceptions import FieldDoesNotExist, PermissionDenied
from django.forms import CheckboxSelectMultiple
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy

from wagtail import hooks
from wagtail.admin.filters import (
    DateRangePickerWidget,
    RelatedFilterMixin,
    WagtailFilterSet,
)
from wagtail.admin.search import SearchArea
from wagtail.admin.ui.menus import MenuItem
from wagtail.admin.ui.tables import (
    BooleanColumn,
    BulkActionsCheckboxColumn,
    Column,
    DateColumn,
    TitleColumn,
)
from wagtail.admin.utils import get_user_display_name
from wagtail.admin.views import generic
from wagtail.admin.viewsets.model import ModelViewSet
from wagtail.admin.widgets.boolean_radio_select import BooleanRadioSelect
from wagtail.admin.widgets.button import (
    BaseButton,
    Button,
    ButtonWithDropdown,
)
from wagtail.compat import AUTH_USER_APP_LABEL, AUTH_USER_MODEL_NAME
from wagtail.search import index
from wagtail.users.forms import UserCreationForm, UserEditForm
from wagtail.users.utils import user_can_delete_user

User = get_user_model()

# Typically we would check the permission 'auth.change_user' (and 'auth.add_user' /
# 'auth.delete_user') for user management actions, but this may vary according to
# the AUTH_USER_MODEL setting. These are no longer used in the codebase in favour
# of ModelPermissionPolicy, but are kept here for backwards compatibility.
add_user_perm = f"{AUTH_USER_APP_LABEL}.add_{AUTH_USER_MODEL_NAME.lower()}"
change_user_perm = "{}.change_{}".format(
    AUTH_USER_APP_LABEL, AUTH_USER_MODEL_NAME.lower()
)
delete_user_perm = "{}.delete_{}".format(
    AUTH_USER_APP_LABEL, AUTH_USER_MODEL_NAME.lower()
)


class UserColumn(TitleColumn):
    cell_template_name = "wagtailusers/users/user_cell.html"


class GroupFilter(RelatedFilterMixin, django_filters.ModelMultipleChoiceFilter):
    pass


class UserFilterSet(WagtailFilterSet):
    is_superuser = django_filters.BooleanFilter(
        label=gettext_lazy("Administrator"),
        widget=BooleanRadioSelect,
    )
    last_login = django_filters.DateFromToRangeFilter(
        label=gettext_lazy("Last login"),
        widget=DateRangePickerWidget,
    )

    def __init__(
        self, data=None, queryset=None, *, request=None, prefix=None, is_searching=False
    ):
        super().__init__(data, queryset, request=request, prefix=prefix)
        try:
            self._meta.model._meta.get_field("is_active")
        except FieldDoesNotExist:
            pass
        else:
            self.filters["is_active"] = django_filters.BooleanFilter(
                field_name="is_active",
                label=gettext_lazy("Active"),
                widget=BooleanRadioSelect,
            )
            self.filters.move_to_end("is_active", last=False)

        self.filters["group"] = GroupFilter(
            field_name="groups",
            queryset=Group.objects.all(),
            label=gettext_lazy("Group"),
            use_subquery=is_searching,
            widget=CheckboxSelectMultiple,
        )

    class Meta:
        model = User
        fields = []


class IndexView(generic.IndexView):
    """
    Lists the users for management within the admin.
    """

    template_name = "wagtailusers/users/index.html"
    results_template_name = "wagtailusers/users/index_results.html"
    add_item_label = gettext_lazy("Add a user")
    context_object_name = "users"
    page_title = gettext_lazy("Users")
    show_other_searches = True

    @cached_property
    def columns(self):
        _UserColumn = self._get_title_column_class(UserColumn)
        return [
            BulkActionsCheckboxColumn("bulk_actions", obj_type="user"),
            _UserColumn(
                "name",
                accessor=lambda u: get_user_display_name(u),
                label=gettext_lazy("Name"),
                sort_key="name"
                if self.model_fields.issuperset({"first_name", "last_name"})
                else None,
                get_url=self.get_edit_url,
                classname="name",
            ),
            Column(
                self.model.USERNAME_FIELD,
                accessor="get_username",
                label=gettext_lazy("Username"),
                sort_key=self.model.USERNAME_FIELD,
                classname="username",
                width="20%",
            ),
            Column(
                "is_superuser",
                accessor=lambda u: gettext_lazy("Admin") if u.is_superuser else None,
                label=gettext_lazy("Access level"),
                sort_key="is_superuser",
                classname="level",
                width="10%",
            ),
            BooleanColumn(
                "is_active",
                label=gettext_lazy("Active"),
                sort_key="is_active" if "is_active" in self.model_fields else None,
                classname="status",
                width="10%",
            ),
            DateColumn(
                "last_login",
                label=gettext_lazy("Last login"),
                sort_key="last_login",
                classname="last-login",
                width="15%",
            ),
        ]

    @cached_property
    def model_fields(self):
        return {f.name for f in User._meta.get_fields()}

    @cached_property
    def search_fields(self):
        # Use search_fields from the model if we're using Wagtail search
        if index.class_is_indexed(User) and self.search_backend_name:
            return None
        return self.model_fields & {"username", "first_name", "last_name", "email"}

    def get_filterset_kwargs(self):
        kwargs = super().get_filterset_kwargs()
        kwargs["is_searching"] = self.is_searching
        return kwargs

    def get_delete_url(self, instance):
        if user_can_delete_user(self.request.user, instance):
            return super().get_delete_url(instance)

    def get_list_buttons(self, instance):
        more_buttons = []
        list_buttons = []

        buttons = self.get_list_more_buttons(instance)
        for hook in hooks.get_hooks("register_user_listing_buttons"):
            buttons.extend(hook(user=instance, request_user=self.request.user))

        for button in buttons:
            if isinstance(button, BaseButton) and not button.allow_in_dropdown:
                # If the button is not allowed in a dropdown menu, add it to
                # the top-level list of buttons
                list_buttons.append(button)
            elif isinstance(button, MenuItem):
                # Allow simple MenuItem instances to be passed in directly
                if button.is_shown(self.request.user):
                    more_buttons.append(Button.from_menu_item(button))
            elif button.show:
                # Otherwise, add it to the default "More" dropdown
                more_buttons.append(button)

        list_buttons.append(
            ButtonWithDropdown(
                buttons=more_buttons,
                icon_name="dots-horizontal",
                attrs={
                    "aria-label": _("More options for '%(title)s'")
                    % {"title": str(instance)},
                },
            )
        )

        return list_buttons

    def get_base_queryset(self):
        users = User._default_manager.all()

        if "wagtail_userprofile" in self.model_fields:
            users = users.select_related("wagtail_userprofile")

        return users

    def order_queryset(self, queryset):
        if self.ordering == "name":
            return queryset.order_by("last_name", "first_name")
        if self.ordering == "-name":
            return queryset.order_by("-last_name", "-first_name")
        return super().order_queryset(queryset)


class CreateView(generic.CreateView):
    """
    Provide the ability to create a user within the admin.
    """

    success_message = gettext_lazy("User '%(object)s' created.")
    page_title = gettext_lazy("Add user")

    def run_before_hook(self):
        return self.run_hook(
            "before_create_user",
            self.request,
        )

    def run_after_hook(self):
        return self.run_hook(
            "after_create_user",
            self.request,
            self.object,
        )


class EditView(generic.EditView):
    """
    Provide the ability to edit a user within the admin.
    """

    success_message = gettext_lazy("User '%(object)s' updated.")
    error_message = gettext_lazy("The user could not be saved due to errors.")
    context_object_name = "user"

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.object = self.get_object()
        self.can_delete = user_can_delete_user(request.user, self.object)
        self.editing_self = request.user == self.object

    def save_instance(self):
        instance = super().save_instance()
        if self.object == self.request.user and "password1" in self.form.changed_data:
            # User is changing their own password; need to update their session hash
            update_session_auth_hash(self.request, self.object)
        return instance

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs.update(
            {
                "editing_self": self.editing_self,
            }
        )
        return kwargs

    def run_before_hook(self):
        return self.run_hook(
            "before_edit_user",
            self.request,
            self.object,
        )

    def run_after_hook(self):
        return self.run_hook(
            "after_edit_user",
            self.request,
            self.object,
        )

    def get_page_subtitle(self):
        return get_user_display_name(self.object)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["can_delete"] = self.can_delete
        return context


class DeleteView(generic.DeleteView):
    """
    Provide the ability to delete a user within the admin.
    """

    page_title = gettext_lazy("Delete user")
    success_message = gettext_lazy("User '%(object)s' deleted.")
    context_object_name = "user"

    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()
        if not user_can_delete_user(self.request.user, self.object):
            raise PermissionDenied
        return super().dispatch(request, *args, **kwargs)

    def run_before_hook(self):
        return self.run_hook(
            "before_delete_user",
            self.request,
            self.object,
        )

    def run_after_hook(self):
        return self.run_hook(
            "after_delete_user",
            self.request,
            self.object,
        )


class HistoryView(generic.HistoryView):
    def get_page_subtitle(self):
        return get_user_display_name(self.object)


class UserViewSet(ModelViewSet):
    icon = "user"
    model = User
    ordering = "name"
    add_to_reference_index = False
    filterset_class = UserFilterSet
    menu_name = "users"
    menu_label = gettext_lazy("Users")
    menu_order = 600
    add_to_settings_menu = True

    index_view_class = IndexView
    add_view_class = CreateView
    edit_view_class = EditView
    delete_view_class = DeleteView
    history_view_class = HistoryView

    template_prefix = "wagtailusers/users/"

    def get_form_class(self, for_update=False):
        if for_update:
            return UserEditForm
        return UserCreationForm

    @cached_property
    def search_area_class(self):
        class UsersSearchArea(SearchArea):
            def is_shown(search_area, request):
                return self.permission_policy.user_has_any_permission(
                    request.user, {"add", "change", "delete"}
                )

        return UsersSearchArea

    def get_search_area(self):
        return self.search_area_class(
            gettext_lazy("Users"),
            reverse(self.get_url_name("index")),
            name="users",
            icon_name="user",
            order=600,
        )

    def register_search_area(self):
        hooks.register("register_admin_search_area", self.get_search_area)

    def on_register(self):
        super().on_register()
        self.register_search_area()



from django.test import TestCase
from django.urls import reverse

from wagtail.test.utils import WagtailTestUtils
from wagtail.users.models import UserProfile


class TestDismissiblesView(WagtailTestUtils, TestCase):
    def setUp(self):
        self.user = self.login()
        self.profile = UserProfile.get_for_user(self.user)
        self.url = reverse("wagtailadmin_dismissibles")

    def test_get_initial(self):
        response = self.client.get(self.url)
        self.profile.refresh_from_db()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {})
        self.assertEqual(self.user.wagtail_userprofile.dismissibles, {})

    def test_patch_valid(self):
        response = self.client.patch(
            self.url, data={"foo": "bar"}, content_type="application/json"
        )
        self.profile.refresh_from_db()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {"foo": "bar"})
        self.assertEqual(self.user.wagtail_userprofile.dismissibles, {"foo": "bar"})

    def test_patch_invalid(self):
        response = self.client.patch(
            self.url, data="invalid", content_type="application/json"
        )
        self.profile.refresh_from_db()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(self.user.wagtail_userprofile.dismissibles, {})

    def test_post(self):
        # The view only accepts GET and PATCH
        response = self.client.post(self.url, data={"foo": "bar"})
        self.profile.refresh_from_db()
        self.assertEqual(response.status_code, 405)
        self.assertEqual(self.user.wagtail_userprofile.dismissibles, {})

    def test_get_without_userprofile(self):
        # GET should work even if the user doesn't have a UserProfile,
        # but it shouldn't create one
        self.profile.delete()
        response = self.client.get(self.url)
        self.user.refresh_from_db()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {})
        self.assertIsNone(getattr(self.user, "wagtail_userprofile", None))

    def test_patch_without_userprofile(self):
        # PATCH should work even if the user doesn't have a UserProfile,
        # in which case it should create one
        self.profile.delete()
        response = self.client.patch(
            self.url, data={"foo": "bar"}, content_type="application/json"
        )
        self.user.refresh_from_db()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {"foo": "bar"})
        self.assertEqual(self.user.wagtail_userprofile.dismissibles, {"foo": "bar"})



"""
Check that all changes to Wagtail models have had migrations created. If there
are outstanding model changes that need migrations, fail the tests.
"""

from django.apps import apps
from django.db.migrations.autodetector import MigrationAutodetector
from django.db.migrations.loader import MigrationLoader
from django.db.migrations.questioner import MigrationQuestioner
from django.db.migrations.state import ProjectState
from django.test import TestCase


class TestForMigrations(TestCase):
    def test__migrations(self):
        app_labels = {
            app.label
            for app in apps.get_app_configs()
            if app.name.split(".")[0] == "wagtail"
        }
        for app_label in app_labels:
            apps.get_app_config(app_label.split(".")[-1])
        loader = MigrationLoader(None, ignore_no_migrations=True)

        conflicts = {
            (app_label, conflict)
            for app_label, conflict in loader.detect_conflicts().items()
            if app_label in app_labels
        }

        if conflicts:
            name_str = "; ".join(
                "{} in {}".format(", ".join(names), app)
                for app, names in conflicts.items()
            )
            self.fail("Conflicting migrations detected (%s)." % name_str)

        autodetector = MigrationAutodetector(
            loader.project_state(),
            ProjectState.from_apps(apps),
            MigrationQuestioner(specified_apps=app_labels, dry_run=True),
        )

        changes = autodetector.changes(
            graph=loader.graph,
            trim_to_apps=app_labels or None,
            convert_apps=app_labels or None,
        )

        if changes:
            migrations = "\n".join(
                "  {migration}\n{changes}".format(
                    migration=migration,
                    changes="\n".join(
                        f"    {operation.describe()}"
                        for operation in migration.operations
                    ),
                )
                for (_, migrations) in changes.items()
                for migration in migrations
            )

            self.fail("Model changes with no migrations detected:\n%s" % migrations)



# Generated by Django 3.0.3 on 2021-04-19 13:54

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import modelcluster.fields
from wagtail.models import COMMENTS_RELATION_NAME


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("wagtailcore", "0061_change_promote_tab_helpt_text_and_verbose_names"),
    ]

    operations = [
        migrations.CreateModel(
            name="Comment",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("text", models.TextField()),
                ("contentpath", models.TextField()),
                ("position", models.TextField(blank=True)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("resolved_at", models.DateTimeField(blank=True, null=True)),
                (
                    "page",
                    modelcluster.fields.ParentalKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name=COMMENTS_RELATION_NAME,
                        to="wagtailcore.Page",
                    ),
                ),
                (
                    "resolved_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="comments_resolved",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "revision_created",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="created_comments",
                        to="wagtailcore.PageRevision",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name=COMMENTS_RELATION_NAME,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "comment",
                "verbose_name_plural": "comments",
            },
        ),
        migrations.CreateModel(
            name="CommentReply",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("text", models.TextField()),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "comment",
                    modelcluster.fields.ParentalKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="replies",
                        to="wagtailcore.Comment",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="comment_replies",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "comment reply",
                "verbose_name_plural": "comment replies",
            },
        ),
        migrations.CreateModel(
            name="PageSubscription",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("comment_notifications", models.BooleanField()),
                (
                    "page",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="subscribers",
                        to="wagtailcore.Page",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="page_subscriptions",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "unique_together": {("page", "user")},
            },
        ),
    ]



from django.urls import path

from wagtail import hooks
from wagtail.api.v2.router import WagtailAPIRouter

from .views import PagesAdminAPIViewSet

admin_api = WagtailAPIRouter("wagtailadmin_api")
admin_api.register_endpoint("pages", PagesAdminAPIViewSet)

for fn in hooks.get_hooks("construct_admin_api"):
    fn(admin_api)

urlpatterns = [
    path("main/", admin_api.urls),
]



from django.test import TestCase
from django.urls import reverse

from wagtail.test.utils import WagtailTestUtils


class TestBulkActionDispatcher(WagtailTestUtils, TestCase):
    def setUp(self):
        # Login
        self.user = self.login()

    def test_bulk_action_invalid_action(self):
        url = reverse(
            "wagtail_bulk_action",
            args=(
                "wagtailcore",
                "page",
                "ships",
            ),
        )
        response = self.client.get(url)
        self.assertEqual(response.status_code, 404)

    def test_bulk_action_invalid_model(self):
        url = reverse(
            "wagtail_bulk_action",
            args=(
                "doesnotexist",
                "doesnotexist",
                "doesnotexist",
            ),
        )
        response = self.client.get(url)
        self.assertEqual(response.status_code, 404)



from django.forms import CheckboxInput, HiddenInput
from django.test import TestCase, override_settings

from wagtail.contrib.simple_translation.forms import SubmitTranslationForm
from wagtail.models import Locale, Page
from wagtail.test.i18n.models import TestPage
from wagtail.test.utils import WagtailTestUtils


@override_settings(
    LANGUAGES=[
        ("en", "English"),
        ("fr", "French"),
        ("de", "German"),
    ],
    WAGTAIL_CONTENT_LANGUAGES=[
        ("en", "English"),
        ("fr", "French"),
        ("de", "German"),
    ],
)
class TestSubmitPageTranslation(WagtailTestUtils, TestCase):
    def setUp(self):
        self.en_locale = Locale.objects.first()
        self.fr_locale = Locale.objects.create(language_code="fr")
        self.de_locale = Locale.objects.create(language_code="de")

        self.en_homepage = Page.objects.get(depth=2)
        self.fr_homepage = self.en_homepage.copy_for_translation(self.fr_locale)
        self.de_homepage = self.en_homepage.copy_for_translation(self.de_locale)

        self.en_blog_index = TestPage(title="Blog", slug="blog")
        self.en_homepage.add_child(instance=self.en_blog_index)

        self.en_blog_post = TestPage(title="Blog post", slug="blog-post")
        self.en_blog_index.add_child(instance=self.en_blog_post)

    def test_include_subtree(self):
        form = SubmitTranslationForm(instance=self.en_blog_post)
        self.assertIsInstance(form.fields["include_subtree"].widget, HiddenInput)

        form = SubmitTranslationForm(instance=self.en_blog_index)
        self.assertIsInstance(form.fields["include_subtree"].widget, CheckboxInput)
        self.assertEqual(
            form.fields["include_subtree"].label, "Include subtree (1 page)"
        )

        form = SubmitTranslationForm(instance=self.en_homepage)
        self.assertEqual(
            form.fields["include_subtree"].label, "Include subtree (2 pages)"
        )

    def test_locales_queryset(self):
        # Homepage is translated to all locales.
        form = SubmitTranslationForm(instance=self.en_homepage)
        self.assertEqual(
            list(
                form.fields["locales"].queryset.values_list("language_code", flat=True)
            ),
            [],
        )
        # Blog index can be translated to `de` and `fr`.
        form = SubmitTranslationForm(instance=self.en_blog_index)
        self.assertEqual(
            list(
                form.fields["locales"].queryset.values_list("language_code", flat=True)
            ),
            ["de", "fr"],
        )
        # Blog post can be translated to `de` and `fr`.
        form = SubmitTranslationForm(instance=self.en_blog_post)
        self.assertEqual(
            list(
                form.fields["locales"].queryset.values_list("language_code", flat=True)
            ),
            ["de", "fr"],
        )

    def test_select_all(self):
        form = SubmitTranslationForm(instance=self.en_homepage)
        # Homepage is translated to all locales.
        self.assertIsInstance(form.fields["select_all"].widget, HiddenInput)

        form = SubmitTranslationForm(instance=self.en_blog_index)
        # Blog post can be translated to `de` and `fr`.
        self.assertIsInstance(form.fields["select_all"].widget, CheckboxInput)

    def test_locale_disabled(self):
        form = SubmitTranslationForm(instance=self.en_blog_post)
        # The parent (blog_index) is translated to English.
        # German and French are disabled.
        self.assertEqual(
            list(form.fields["locales"].widget.disabled_values),
            [self.de_locale.id, self.fr_locale.id],
        )
        label = f"""
        <label class="disabled">
            <input type="checkbox" name="None" value="{self.de_locale.id}" disabled>
            German
        </label>
        """
        self.assertInHTML(label, form.fields["locales"].widget.render(None, None))

    def test_locale_help_text(self):
        # German and French are disabled.
        # The help_text is plural
        form = SubmitTranslationForm(instance=self.en_blog_post)
        help_text = f"""
            Some locales are disabled because some parent pages are not translated.
            <br>
            <a href="/admin/translation/submit/page/{self.en_blog_index.id}/">
                Translate the parent pages.
            </a>
        """
        self.assertHTMLEqual(form.fields["locales"].help_text, help_text)

        # Add German translation
        self.en_blog_index.copy_for_translation(self.de_locale)
        # French is disabled.
        # The help_text is singular.
        form = SubmitTranslationForm(instance=self.en_blog_post)
        help_text = f"""
            A locale is disabled because a parent page is not translated.
            <br>
            <a href="/admin/translation/submit/page/{self.en_blog_index.id}/">
                Translate the parent page.
            </a>
        """
        self.assertHTMLEqual(form.fields["locales"].help_text, help_text)

    def test_hide_submit(self):
        # German and French are disabled.
        # There are no other pages to be translated.
        # Submit is hidden.
        form = SubmitTranslationForm(instance=self.en_blog_post)
        self.assertFalse(form.show_submit)
        # A parent is translated
        self.en_blog_index.copy_for_translation(self.de_locale)
        form = SubmitTranslationForm(instance=self.en_blog_post)
        self.assertTrue(form.show_submit)



from django import forms
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db import transaction
from django.db.models import Q
from django.utils.functional import cached_property
from django.utils.text import capfirst
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy

from wagtail.admin import widgets
from wagtail.admin.forms import WagtailAdminModelForm
from wagtail.admin.forms.formsets import BaseFormSetMixin
from wagtail.admin.panels import FieldPanel, InlinePanel, ObjectList
from wagtail.admin.widgets.workflows import AdminTaskChooser
from wagtail.coreutils import get_content_type_label, get_model_string
from wagtail.models import Page, Task, Workflow, WorkflowContentType, WorkflowPage
from wagtail.snippets.models import get_workflow_enabled_models


class TaskChooserSearchForm(forms.Form):
    q = forms.CharField(
        label=gettext_lazy("Search term"), widget=forms.TextInput(), required=False
    )

    def __init__(self, *args, task_type_choices=None, **kwargs):
        placeholder = kwargs.pop("placeholder", _("Search…"))
        super().__init__(*args, **kwargs)
        self.fields["q"].widget.attrs = {"placeholder": placeholder}

        # Add task type filter if there is more than one task type option
        if task_type_choices and len(task_type_choices) > 1:
            self.fields["task_type"] = forms.ChoiceField(
                choices=(
                    # Append an "All types" choice to the beginning
                    [(None, _("All types"))]
                    # The task type choices that are passed in use the models as values, we need
                    # to convert these to something that can be represented in HTML
                    + [
                        (get_model_string(model), verbose_name)
                        for model, verbose_name in task_type_choices
                    ]
                ),
                required=False,
            )

        # Save a mapping of task_type values back to the model that we can reference later
        self.task_type_choices = {
            get_model_string(model): model for model, verbose_name in task_type_choices
        }

    def is_searching(self):
        """
        Returns True if the user typed a search query
        """
        return self.is_valid() and bool(self.cleaned_data.get("q"))

    @cached_property
    def task_model(self):
        """
        Returns the selected task model.

        This looks for the task model in the following order:
         1) If there's only one task model option, return it
         2) If a task model has been selected, return it
         3) Return the generic Task model
        """
        models = list(self.task_type_choices.values())
        if len(models) == 1:
            return models[0]

        elif self.is_valid():
            model_name = self.cleaned_data.get("task_type")
            if model_name and model_name in self.task_type_choices:
                return self.task_type_choices[model_name]

        return Task

    def specific_task_model_selected(self):
        return self.task_model is not Task


class WorkflowPageForm(forms.ModelForm):
    page = forms.ModelChoiceField(
        queryset=Page.objects.all(),
        widget=widgets.AdminPageChooser(target_models=[Page], can_choose_root=True),
    )

    class Meta:
        model = WorkflowPage
        fields = ["page"]

    def clean(self):
        page = self.cleaned_data.get("page")
        try:
            existing_workflow = page.workflowpage.workflow
            if not self.errors and existing_workflow != self.cleaned_data["workflow"]:
                # If the form has no errors, Page has an existing Workflow assigned, that Workflow is not
                # the selected Workflow, and overwrite_existing is not True, add a new error. This should be used to
                # trigger the confirmation message in the view. This is why this error is only added if there are no
                # other errors - confirmation should be the final step.
                self.add_error(
                    "page",
                    ValidationError(
                        _(
                            "This page already has workflow '%(workflow_name)s' assigned."
                        )
                        % {"workflow_name": existing_workflow},
                        code="existing_workflow",
                    ),
                )
        except AttributeError:
            pass

    def save(self, commit=False):
        page = self.cleaned_data["page"]

        if commit:
            WorkflowPage.objects.update_or_create(
                page=page,
                defaults={"workflow": self.cleaned_data["workflow"]},
            )


class BaseWorkflowPagesFormSet(BaseFormSetMixin, forms.BaseInlineFormSet):
    def clean(self):
        """Checks that no two forms refer to the same page object"""
        if any(self.errors):
            # Don't bother validating the formset unless each form is valid on its own
            return

        pages = [
            form.cleaned_data["page"]
            for form in self.forms
            # need to check for presence of 'page' in cleaned_data,
            # because a completely blank form passes validation
            if form not in self.deleted_forms and "page" in form.cleaned_data
        ]
        if len(set(pages)) != len(pages):
            # pages list contains duplicates
            raise forms.ValidationError(
                _("You cannot assign this workflow to the same page multiple times.")
            )


class WorkflowContentTypeForm(forms.Form):
    class ContentTypeMultipleChoiceField(forms.ModelMultipleChoiceField):
        def label_from_instance(self, obj):
            return get_content_type_label(obj)

    class CheckboxSelectMultiple(forms.CheckboxSelectMultiple):
        """Custom CheckboxSelectMultiple widget that renders errors for each content type ID"""

        option_template_name = (
            "wagtailadmin/workflows/includes/workflow_content_types_checkbox.html"
        )

        def get_errors_by_id(self, errors):
            errors_by_id = {}
            for error in errors.as_data():
                ct_id = error.params and error.params.get("content_type_id")
                errors_by_id.setdefault(ct_id, []).append(error)
            return errors_by_id

        def render_with_errors(
            self, name, value, attrs=None, renderer=None, errors=None
        ):
            context = {
                **self.get_context(name, value, attrs),
                "errors_by_id": self.get_errors_by_id(errors),
            }
            return self._render(self.template_name, context, renderer)

    content_types = ContentTypeMultipleChoiceField(
        queryset=ContentType.objects.none(),
        widget=CheckboxSelectMultiple(),
        required=False,
    )

    def __init__(self, *args, workflow=None, **kwargs):
        self.workflow = workflow
        if workflow and "initial" not in kwargs:
            kwargs["initial"] = {"content_types": workflow.workflow_content_types.all()}

        super().__init__(*args, **kwargs)

        # Start with an always-false query, as Django can optimise it by
        # returning an empty queryset without running any database queries.
        workflow_enabled_q = Q(pk__in=[])

        # Then union the query for each workflow-enabled model.
        for model in get_workflow_enabled_models():
            workflow_enabled_q |= Q(
                app_label=model._meta.app_label, model=model._meta.model_name
            )

        self.fields["content_types"].queryset = ContentType.objects.filter(
            workflow_enabled_q
        )

    def clean(self):
        content_types = self.cleaned_data.get("content_types")
        if not content_types:
            return

        existing_assignments = WorkflowContentType.objects.filter(
            content_type__in=content_types,
            workflow__active=True,
        ).exclude(workflow=self.workflow)
        for assignment in existing_assignments:
            self.add_error(
                "content_types",
                ValidationError(
                    _(
                        "Snippet '%(content_type)s' already has workflow '%(workflow_name)s' assigned."
                    )
                    % {
                        "content_type": capfirst(assignment.content_type.name),
                        "workflow_name": assignment.workflow,
                    },
                    code="existing_workflow_content_type",
                    params={"content_type_id": assignment.content_type_id},
                ),
            )

    def save(self, commit=True):
        if not commit:
            return

        content_types = self.cleaned_data["content_types"]

        with transaction.atomic():
            # Remove any content types that are no longer selected
            WorkflowContentType.objects.filter(workflow=self.workflow).exclude(
                content_type__in=content_types
            ).delete()

            # Add any new content types, ignoring conflicts with existing ones
            # to avoid additional query for existing content types
            objects = [
                WorkflowContentType(workflow=self.workflow, content_type=ct)
                for ct in content_types
            ]
            WorkflowContentType.objects.bulk_create(objects, ignore_conflicts=True)


WorkflowPagesFormSet = forms.inlineformset_factory(
    Workflow,
    WorkflowPage,
    form=WorkflowPageForm,
    formset=BaseWorkflowPagesFormSet,
    extra=1,
    can_delete=True,
    fields=["page"],
)


class BaseTaskForm(forms.ModelForm):
    pass


def get_task_form_class(task_model, for_edit=False):
    """
    Generates a form class for the given task model.

    If the form is to edit an existing task, set for_edit to True. This applies
    the readonly restrictions on fields defined in admin_form_readonly_on_edit_fields.
    """
    fields = task_model.admin_form_fields

    form_class = forms.modelform_factory(
        task_model,
        form=BaseTaskForm,
        fields=fields,
        widgets=getattr(task_model, "admin_form_widgets", {}),
    )

    if for_edit:
        for field_name in getattr(task_model, "admin_form_readonly_on_edit_fields", []):
            if field_name not in form_class.base_fields:
                raise ImproperlyConfigured(
                    "`%s.admin_form_readonly_on_edit_fields` contains the field "
                    "'%s' that doesn't exist. Did you forget to add "
                    "it to `%s.admin_form_fields`?"
                    % (task_model.__name__, field_name, task_model.__name__)
                )

            form_class.base_fields[field_name].disabled = True

    return form_class


def get_workflow_edit_handler():
    """
    Returns an edit handler which provides the "name" and "tasks" fields for workflow.
    """
    # Note. It's a bit of a hack that we use edit handlers here. Ideally, it should be
    # made easier to reuse the inline panel templates for any formset.
    # Since this form is internal, we're OK with this for now. We might want to revisit
    # this decision later if we decide to allow custom fields on Workflows.

    panels = [
        FieldPanel("name", heading=_("Give your workflow a name")),
        InlinePanel(
            "workflow_tasks",
            [
                FieldPanel("task", widget=AdminTaskChooser(show_clear_link=False)),
            ],
            heading=_("Add tasks to your workflow"),
            label=_("task"),
            icon="thumbtack",
        ),
    ]
    edit_handler = ObjectList(panels, base_form_class=WagtailAdminModelForm)
    return edit_handler.bind_to_model(Workflow)



import django_filters
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db.models import Count, F, OuterRef, Q, Subquery
from django.utils.translation import gettext_lazy as _

from wagtail.admin.filters import WagtailFilterSet
from wagtail.admin.views.reports import ReportView
from wagtail.coreutils import get_content_languages
from wagtail.models import Page, Site, get_page_models
from wagtail.permissions import page_permission_policy


def _get_locale_choices():
    return list(get_content_languages().items())


def _get_site_choices():
    """Tuples of (site root page path, site display name) for all sites in project."""
    choices = [
        (site.root_page.path, str(site))
        for site in Site.objects.all().select_related("root_page")
    ]
    return choices


def _annotate_last_edit_info(queryset, language_code, site_root_path):
    latest_edited_page_filter_kwargs = {}
    page_count_filter_kwargs = {}
    if language_code:
        latest_edited_page_filter_kwargs["locale__language_code"] = language_code
        page_count_filter_kwargs["pages__locale__language_code"] = language_code
    if site_root_path:
        latest_edited_page_filter_kwargs["path__startswith"] = site_root_path
        page_count_filter_kwargs["pages__path__startswith"] = site_root_path

    latest_edited_page = Page.objects.filter(
        content_type=OuterRef("pk"), **latest_edited_page_filter_kwargs
    ).order_by(F("latest_revision_created_at").desc(nulls_last=True), "title", "-pk")

    queryset = queryset.annotate(
        count=Count("pages", filter=Q(**page_count_filter_kwargs)),
        last_edited_page_id=Subquery(latest_edited_page.values("pk")[:1]),
    )

    return queryset


class LocaleFilter(django_filters.ChoiceFilter):
    def filter(self, qs, language_code):
        if language_code:
            return qs.filter(pages__locale__language_code=language_code)
        return qs


class SiteFilter(django_filters.ChoiceFilter):
    def filter(self, qs, site_root_path):
        # Value passed will be the site root page path
        # To check if a page is in a site, we check if the page path starts with the
        # site's root page path
        if site_root_path:
            return qs.filter(pages__path__startswith=site_root_path)
        return qs


class PageTypesUsageReportFilterSet(WagtailFilterSet):
    page_locale = LocaleFilter(
        label=_("Locale"),
        choices=_get_locale_choices,
        empty_label=None,
        null_label=_("All"),
        null_value=None,
    )
    site = SiteFilter(
        label=_("Site"),
        choices=_get_site_choices,
        empty_label=None,
        null_label=_("All"),
        null_value=None,
    )

    class Meta:
        model = ContentType
        fields = ["page_locale", "site"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sites = {
            site.pk: site for site in Site.objects.all().prefetch_related("root_page")
        }
        self.sites_filter_enabled = True
        if len(self.sites) == 1:
            # If there is only one site, we don't need to show the site filter
            self.sites_filter_enabled = False
            del self.filters["site"]


class PageTypesUsageReportView(ReportView):
    results_template_name = "wagtailadmin/reports/page_types_usage_results.html"
    page_title = _("Page types usage")
    header_icon = "doc-empty-inverse"
    filterset_class = PageTypesUsageReportFilterSet
    index_url_name = "wagtailadmin_reports:page_types_usage"
    index_results_url_name = "wagtailadmin_reports:page_types_usage_results"
    permission_policy = page_permission_policy
    any_permission_required = ["add", "change", "publish"]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.page_models = [model.__name__.lower() for model in get_page_models()]
        self.i18n_enabled = getattr(settings, "WAGTAIL_I18N_ENABLED", False)

    def decorate_paginated_queryset(self, page_types):
        pages_mapping = Page.objects.specific().in_bulk(
            obj.last_edited_page_id for obj in page_types if obj.last_edited_page_id
        )

        for item in page_types:
            item.last_edited_page = (
                pages_mapping[item.last_edited_page_id]
                if item.last_edited_page_id
                else None
            )

        return page_types

    def get_queryset(self):
        queryset = ContentType.objects.filter(model__in=self.page_models)

        page_content_type = ContentType.objects.get_for_model(Page)
        has_pages = Page.objects.filter(
            depth__gt=1, content_type=page_content_type
        ).exists()
        if not has_pages:
            # If there are no `wagtailcore.Page` pages, we don't need to
            # show it in the report
            queryset = queryset.exclude(id=page_content_type.id)

        self.queryset = queryset

        queryset = self.filter_queryset(queryset)

        language_code = self.filters.form.cleaned_data.get("page_locale", None)
        site_root_path = self.filters.form.cleaned_data.get("site", None)
        queryset = _annotate_last_edit_info(queryset, language_code, site_root_path)

        queryset = queryset.order_by("-count", "app_label", "model")

        return queryset



# Generated by Django 5.0.14 on 2025-06-02 20:03

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
        ("wagtailcore", "0094_alter_page_locale"),
    ]

    operations = [
        migrations.CreateModel(
            name="GroupSitePermission",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "group",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="site_permissions",
                        to="auth.group",
                        verbose_name="group",
                    ),
                ),
                (
                    "permission",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="auth.permission",
                        verbose_name="permission",
                    ),
                ),
                (
                    "site",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="group_permissions",
                        to="wagtailcore.site",
                        verbose_name="site",
                    ),
                ),
            ],
            options={
                "verbose_name": "group site permission",
                "verbose_name_plural": "group site permissions",
                "unique_together": {("group", "site", "permission")},
            },
        ),
    ]



from django import forms
from django.forms.models import inlineformset_factory
from django.utils.translation import gettext_lazy as _

from wagtail.admin.forms.models import WagtailAdminModelForm
from wagtail.admin.widgets import AdminPageChooser
from wagtail.contrib.search_promotions.models import Query, SearchPromotion


class QueryForm(forms.ModelForm):
    query_string = forms.CharField(
        label=_("Search term(s)/phrase"),
        help_text=_(
            "Enter the full search string to match. An "
            "exact match is required for your Promoted Results to be "
            "displayed, wildcards are NOT allowed."
        ),
        required=True,
    )

    def clean(self):
        # We allow using an existing query string on the CreateView, so we need
        # to skip the unique validation on `query_string`. This can be done by
        # overriding the `clean()` method without calling `super().clean()`:
        # https://docs.djangoproject.com/en/stable/topics/forms/modelforms/#overriding-the-clean-method
        pass

    class Meta:
        model = Query
        fields = ["query_string"]


class SearchPromotionForm(WagtailAdminModelForm):
    sort_order = forms.IntegerField(required=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["page"].widget = AdminPageChooser()

    def clean(self):
        cleaned_data = super().clean()

        # Use the raw value instead of from form.cleaned_data so we don't
        # consider an invalid field as empty. For example, leaving the
        # page field empty and entering an invalid external_link_url
        # shouldn't raise an error about needing to enter a page or URL,
        # since the user *has* entered (or tried to enter) a URL.
        page = self["page"].value()
        external_link_url = self["external_link_url"].value()
        external_link_text = self["external_link_text"].value()

        # Must supply a page or external_link_url (but not both)
        if page is None:
            if external_link_url:
                # if an external_link_url is supplied,
                # then external_link_text is also required
                if not external_link_text:
                    self.add_error(
                        "external_link_text",
                        forms.ValidationError(
                            _(
                                "You must enter an external link text if you enter an external link URL."
                            )
                        ),
                    )
            else:
                self.add_error(
                    None,
                    forms.ValidationError(
                        _("You must recommend a page OR an external link.")
                    ),
                )
        elif external_link_url:
            self.add_error(
                None,
                forms.ValidationError(
                    _("Please only select a page OR enter an external link.")
                ),
            )

        return cleaned_data

    class Meta(WagtailAdminModelForm.Meta):
        model = SearchPromotion
        fields = (
            "query",
            "page",
            "external_link_url",
            "external_link_text",
            "description",
        )

        widgets = {
            "description": forms.Textarea(attrs={"rows": 3}),
        }


SearchPromotionsFormSetBase = inlineformset_factory(
    Query,
    SearchPromotion,
    form=SearchPromotionForm,
    can_order=True,
    can_delete=True,
    extra=0,
)


class SearchPromotionsFormSet(SearchPromotionsFormSetBase):
    minimum_forms = 1
    minimum_forms_message = _(
        "Please specify at least one recommendation for this search term."
    )

    def add_fields(self, form, *args, **kwargs):
        super().add_fields(form, *args, **kwargs)

        # Hide delete and order fields
        form.fields["DELETE"].widget = forms.HiddenInput()
        form.fields["ORDER"].widget = forms.HiddenInput()

        # Remove query field
        del form.fields["query"]

    def clean(self):
        # Search pick must have at least one recommended page to be valid
        # Check there is at least one non-deleted form.
        non_deleted_forms = self.total_form_count()
        non_empty_forms = 0
        for i in range(0, self.total_form_count()):
            form = self.forms[i]
            if self.can_delete and self._should_delete_form(form):
                non_deleted_forms -= 1
            if not (form.instance.id is None and not form.has_changed()):
                non_empty_forms += 1
        if (
            non_deleted_forms < self.minimum_forms
            or non_empty_forms < self.minimum_forms
        ):
            raise forms.ValidationError(self.minimum_forms_message)



# Generated by Django 3.1.1 on 2020-09-30 10:14

from django.db import migrations

# import the real Collection model (as opposed to the frozen version we'd obtain with
# apps.get_model) so that we can use its fix_tree method. Importing real models in migrations is
# generally avoided since it's liable to generate SQL queries according to the final migrated
# database state which we haven't reached yet - however, in this case it should be safe because
# fix_tree only touches the Treebeard-specific fields via values_list and update, and shouldn't be
# sensitive to schema changes elsewhere.
from wagtail.models import Collection


def apply_collection_ordering(apps, schema_editor):
    Collection.fix_tree(fix_paths=True)


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailcore", "0058_page_alias_of"),
    ]

    operations = [
        migrations.RunPython(apply_collection_ordering, migrations.RunPython.noop),
    ]



from django.core.exceptions import PermissionDenied

from wagtail.log_actions import log


class ConvertAliasPageError(RuntimeError):
    """
    Raised when the page to convert is not an alias.
    """

    pass


class ConvertAliasPagePermissionError(PermissionDenied):
    """
    Raised when the alias page conversion cannot be performed due to insufficient permissions.
    """

    pass


class ConvertAliasPageAction:
    def __init__(self, page, *, log_action="wagtail.convert_alias", user=None):
        self.page = page
        self.log_action = log_action
        self.user = user

    def check(self, skip_permission_checks=False):
        if not self.page.alias_of_id:
            raise ConvertAliasPageError("Page must be an alias to be converted.")

        if (
            not skip_permission_checks
            and self.user
            and not self.page.permissions_for_user(self.user).can_edit()
        ):
            raise ConvertAliasPagePermissionError(
                "You do not have permission to edit this page."
            )

    def _convert_alias(self, page, log_action, user):
        page.alias_of_id = None
        page.save(update_fields=["alias_of_id"], clean=False)

        # Create an initial revision
        revision = page.save_revision(user=user, changed=False, clean=False)

        if page.live:
            page.live_revision = revision
            page.save(update_fields=["live_revision"], clean=False)

        # Log
        if log_action:
            log(
                instance=page,
                action=log_action,
                revision=revision,
                user=user,
                data={
                    "page": {"id": page.id, "title": page.get_admin_display_title()},
                },
            )

        return page

    def execute(self, skip_permission_checks=False):
        self.check(skip_permission_checks=skip_permission_checks)

        return self._convert_alias(self.page, self.log_action, self.user)



# Generated by Django 2.1.7 on 2019-03-04 08:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailembeds", "0004_embed_verbose_name_plural"),
    ]

    operations = [
        migrations.AlterField(
            model_name="embed",
            name="thumbnail_url",
            field=models.URLField(blank=True, max_length=255, null=True),
        ),
    ]



# Generated by Django 4.2.4 on 2023-08-21 10:56

import datetime
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("tests", "0026_featurecompletetoy"),
    ]

    operations = [
        migrations.AddField(
            model_name="featurecompletetoy",
            name="release_date",
            field=models.DateField(default=datetime.date.today),
        ),
    ]



# Generated by Django 5.0.3 on 2024-04-15 22:12

import django.db.models.deletion
import wagtail.blocks
import wagtail.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0035_modelwithcustommanager'),
        ('wagtailcore', '0093_uploadedfile'),
    ]

    operations = [
        migrations.CreateModel(
            name='ComplexDefaultStreamPage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.page')),
                ('body', wagtail.fields.StreamField([('text', wagtail.blocks.CharBlock()), ('rich_text', wagtail.blocks.RichTextBlock()), ('books', wagtail.blocks.StreamBlock([('title', wagtail.blocks.CharBlock()), ('author', wagtail.blocks.CharBlock())]))], default=[('rich_text', '<p>My <i>lovely</i> books</p>'), ('books', [('title', 'The Great Gatsby'), ('author', 'F. Scott Fitzgerald')])])),
            ],
            options={
                'abstract': False,
            },
            bases=('wagtailcore.page',),
        ),
    ]



import uuid

from asgiref.local import Local
from django.utils.functional import LazyObject

from wagtail import hooks
from wagtail.utils.registry import ObjectTypeRegistry


class LogFormatter:
    """
    Defines how to format log messages / comments for a particular action type. Messages that depend on
    log entry data should override format_message / format_comment; static messages can just be set as the
    'message' / 'comment' attribute.

    To be registered with log_registry.register_action.
    """

    label = ""
    message = ""
    comment = ""

    def format_message(self, log_entry):
        return self.message

    def format_comment(self, log_entry):
        return self.comment


_active = Local()


class LogContext:
    """
    Stores data about the environment in which a logged action happens -
    e.g. the active user - to be stored in the log entry for that action.
    """

    def __init__(self, user=None, generate_uuid=True):
        self.user = user
        if generate_uuid:
            self.uuid = uuid.uuid4()
        else:
            self.uuid = None

    def __enter__(self):
        self._old_log_context = getattr(_active, "value", None)
        activate(self)
        return self

    def __exit__(self, type, value, traceback):
        if self._old_log_context:
            activate(self._old_log_context)
        else:
            deactivate()


empty_log_context = LogContext(generate_uuid=False)


def activate(log_context):
    _active.value = log_context


def deactivate():
    del _active.value


def get_active_log_context():
    return getattr(_active, "value", empty_log_context)


class LogActionRegistry:
    """
    A central store for log actions.
    The expected format for registered log actions: Namespaced action, Action label, Action message (or callable)
    """

    def __init__(self):
        # Has the register_log_actions hook been run for this registry?
        self.has_scanned_for_actions = False

        # Holds the formatter objects, keyed by action
        self.formatters = {}

        # Holds a list of action, action label tuples for use in filters
        self.choices = []

        # Tracks which LogEntry model should be used for a given object class
        self.log_entry_models_by_type = ObjectTypeRegistry()

        # All distinct log entry models registered with register_model
        self.log_entry_models = set()

    def scan_for_actions(self):
        if not self.has_scanned_for_actions:
            for fn in hooks.get_hooks("register_log_actions"):
                fn(self)

            self.has_scanned_for_actions = True

    def register_model(self, cls, log_entry_model):
        self.log_entry_models_by_type.register(cls, value=log_entry_model)
        self.log_entry_models.add(log_entry_model)

    def register_action(self, action, *args):
        def register_formatter_class(formatter_cls):
            formatter = formatter_cls()
            self.formatters[action] = formatter
            self.choices.append((action, formatter.label))

        if args:
            # register_action has been invoked as register_action(action, label, message); create a LogFormatter
            # subclass and register that
            label, message = args
            formatter_cls = type(
                "_LogFormatter", (LogFormatter,), {"label": label, "message": message}
            )
            register_formatter_class(formatter_cls)
        else:
            # register_action has been invoked as a @register_action(action) decorator; return the function that
            # will register the class
            return register_formatter_class

    def get_choices(self):
        self.scan_for_actions()
        return self.choices

    def get_formatter(self, log_entry):
        self.scan_for_actions()
        return self.formatters.get(log_entry.action)

    def action_exists(self, action):
        self.scan_for_actions()
        return action in self.formatters

    def get_log_entry_models(self):
        self.scan_for_actions()
        return self.log_entry_models

    def get_action_label(self, action):
        return self.formatters[action].label

    def get_log_model_for_model(self, model):
        self.scan_for_actions()
        return self.log_entry_models_by_type.get_by_type(model)

    def get_log_model_for_instance(self, instance):
        if isinstance(instance, LazyObject):
            # for LazyObject instances such as request.user, ensure we're looking up the real type
            instance = instance._wrapped

        return self.get_log_model_for_model(type(instance))

    def log(self, instance, action, user=None, uuid=None, **kwargs):
        self.scan_for_actions()

        # find the log entry model for the given object type
        log_entry_model = self.get_log_model_for_instance(instance)
        if log_entry_model is None:
            # no logger registered for this object type - silently bail
            return

        user = user or get_active_log_context().user
        uuid = uuid or get_active_log_context().uuid
        return log_entry_model.objects.log_action(
            instance, action, user=user, uuid=uuid, **kwargs
        )

    def get_logs_for_instance(self, instance):
        log_entry_model = self.get_log_model_for_instance(instance)
        if log_entry_model is None:
            # this model has no logs; return an empty queryset of the basic log model
            from wagtail.models import ModelLogEntry

            return ModelLogEntry.objects.none()

        return log_entry_model.objects.for_instance(instance)


registry = LogActionRegistry()


def log(instance, action, **kwargs):
    return registry.log(instance, action, **kwargs)



# Generated by Django 4.0.4 on 2022-05-25 15:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailcore", "0073_page_latest_revision"),
    ]

    operations = [
        migrations.AddField(
            model_name="revision",
            name="object_str",
            field=models.TextField(default=""),
        ),
    ]



# Generated by Django 4.2.5 on 2023-09-26 09:29

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("tests", "0028_fullfeaturedsnippet_some_number"),
    ]

    operations = [
        migrations.AddField(
            model_name="variousondeletemodel",
            name="cascading_toy",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="tests.featurecompletetoy",
            ),
        ),
    ]



from io import StringIO

from django.contrib.auth.models import Permission
from django.core import management
from django.test import TransactionTestCase
from django.urls import reverse
from django.utils.http import urlencode

from wagtail.models import Page
from wagtail.test.testapp.models import EventIndex, SimplePage, SingleEventPage
from wagtail.test.utils import WagtailTestUtils
from wagtail.test.utils.timestamps import local_datetime


class TestPageSearch(WagtailTestUtils, TransactionTestCase):
    fixtures = ["test_empty.json"]

    def setUp(self):
        super().setUp()
        management.call_command(
            "update_index",
            backend_name="default",
            stdout=StringIO(),
            chunk_size=50,
        )
        self.user = self.login()

    def get(self, params=None, url_name="wagtailadmin_pages:search", **extra):
        return self.client.get(reverse(url_name), params or {}, **extra)

    def test_view(self):
        response = self.get()
        self.assertTemplateUsed(response, "wagtailadmin/pages/search.html")
        self.assertEqual(response.status_code, 200)

        with self.assertNumQueries(22):
            self.get()

    def test_search(self):
        # Find root page
        root_page = Page.objects.get(id=2)

        # Create a page
        new_page = root_page.add_child(
            instance=SimplePage(
                title="Hello from Cauldron Lake",
                slug="bright-falls",
                content="It's not a lake, it's an ocean",
                live=True,
                has_unpublished_changes=False,
            )
        )

        response = self.get({"q": "Hello"})
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "wagtailadmin/pages/search.html")
        self.assertEqual(response.context["query_string"], "Hello")
        next_url = urlencode({"next": reverse("wagtailadmin_pages:search")})
        expected_new_page_copy_url = (
            reverse("wagtailadmin_pages:copy", args=(new_page.pk,)) + f"?{next_url}"
        )
        self.assertContains(response, f'href="{expected_new_page_copy_url}"')

    def test_search_searchable_fields(self):
        # Find root page
        root_page = Page.objects.get(id=2)

        # Create a page
        root_page.add_child(
            instance=SimplePage(
                title="Greetings!",
                slug="hello",
                content="good morning",
                live=True,
                has_unpublished_changes=False,
            )
        )

        # Confirm the slug is not being searched
        response = self.get({"q": "hello"})
        self.assertNotContains(response, "There is one matching page")

        # Confirm the title is being searched
        response = self.get({"q": "greetings"})
        self.assertContains(response, "There is one matching page")

    def test_ajax(self):
        # Find root page
        root_page = Page.objects.get(id=2)

        # Create a page
        new_page = root_page.add_child(
            instance=SimplePage(
                title="Hello from Cauldron Lake",
                slug="bright-falls",
                content="It's not a lake, it's an ocean",
                live=True,
                has_unpublished_changes=False,
            )
        )

        response = self.get(
            {"q": "Hello"}, url_name="wagtailadmin_pages:search_results"
        )
        self.assertEqual(response.status_code, 200)
        self.assertTemplateNotUsed(response, "wagtailadmin/pages/search.html")
        self.assertTemplateUsed(response, "wagtailadmin/pages/search_results.html")
        self.assertEqual(response.context["query_string"], "Hello")
        next_url = urlencode({"next": reverse("wagtailadmin_pages:search")})
        expected_new_page_copy_url = (
            reverse("wagtailadmin_pages:copy", args=(new_page.pk,)) + f"?{next_url}"
        )
        self.assertContains(response, f'href="{expected_new_page_copy_url}"')

    def test_pagination(self):
        pages = ["0", "1", "-1", "9999", "Not a page"]
        for page in pages:
            response = self.get({"q": "Hello", "p": page})
            self.assertEqual(response.status_code, 200)
            self.assertTemplateUsed(response, "wagtailadmin/pages/search.html")

    def test_root_can_appear_in_search_results(self):
        response = self.get({"q": "root"})
        self.assertEqual(response.status_code, 200)
        # 'pages' list in the response should contain root
        results = response.context["pages"]
        self.assertTrue(any(r.slug == "root" for r in results))

    def test_search_uses_admin_display_title_from_specific_class(self):
        # SingleEventPage has a custom get_admin_display_title method; explorer should
        # show the custom title rather than the basic database one
        root_page = Page.objects.get(id=2)
        new_event = SingleEventPage(
            title="Lunar event",
            location="the moon",
            audience="public",
            cost="free",
            date_from="2001-01-01",
            latest_revision_created_at=local_datetime(2016, 1, 1),
        )
        root_page.add_child(instance=new_event)
        response = self.get({"q": "lunar"})
        self.assertContains(response, "Lunar event (single event)")

    def test_search_no_perms(self):
        self.user.is_superuser = False
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            )
        )
        self.user.save()
        self.assertRedirects(self.get(), "/admin/")

    def test_search_order_by_title(self):
        root_page = Page.objects.get(id=2)
        new_event = SingleEventPage(
            title="Lunar event",
            location="the moon",
            audience="public",
            cost="free",
            date_from="2001-01-01",
            latest_revision_created_at=local_datetime(2016, 1, 1),
        )
        root_page.add_child(instance=new_event)

        new_event_2 = SingleEventPage(
            title="A Lunar event",
            location="the moon",
            audience="public",
            cost="free",
            date_from="2001-01-01",
            latest_revision_created_at=local_datetime(2016, 1, 1),
        )
        root_page.add_child(instance=new_event_2)

        response = self.get({"q": "Lunar", "ordering": "title"})
        page_ids = [page.id for page in response.context["pages"]]
        self.assertEqual(page_ids, [new_event_2.id, new_event.id])

        response = self.get({"q": "Lunar", "ordering": "-title"})
        page_ids = [page.id for page in response.context["pages"]]
        self.assertEqual(page_ids, [new_event.id, new_event_2.id])

    def test_search_order_by_updated(self):
        root_page = Page.objects.get(id=2)
        new_event = SingleEventPage(
            title="Lunar event",
            location="the moon",
            audience="public",
            cost="free",
            date_from="2001-01-01",
            latest_revision_created_at=local_datetime(2016, 1, 1),
        )
        root_page.add_child(instance=new_event)

        new_event_2 = SingleEventPage(
            title="Lunar event 2",
            location="the moon",
            audience="public",
            cost="free",
            date_from="2001-01-01",
            latest_revision_created_at=local_datetime(2015, 1, 1),
        )
        root_page.add_child(instance=new_event_2)

        response = self.get({"q": "Lunar", "ordering": "latest_revision_created_at"})
        page_ids = [page.id for page in response.context["pages"]]
        self.assertEqual(page_ids, [new_event_2.id, new_event.id])

        response = self.get({"q": "Lunar", "ordering": "-latest_revision_created_at"})
        page_ids = [page.id for page in response.context["pages"]]
        self.assertEqual(page_ids, [new_event.id, new_event_2.id])

    def test_search_order_by_status(self):
        root_page = Page.objects.get(id=2)
        live_event = SingleEventPage(
            title="Lunar event",
            location="the moon",
            audience="public",
            cost="free",
            date_from="2001-01-01",
            latest_revision_created_at=local_datetime(2016, 1, 1),
            live=True,
        )
        root_page.add_child(instance=live_event)

        draft_event = SingleEventPage(
            title="Lunar event",
            location="the moon",
            audience="public",
            cost="free",
            date_from="2001-01-01",
            latest_revision_created_at=local_datetime(2016, 1, 1),
            live=False,
        )
        root_page.add_child(instance=draft_event)

        response = self.get({"q": "Lunar", "ordering": "live"})
        page_ids = [page.id for page in response.context["pages"]]
        self.assertEqual(page_ids, [draft_event.id, live_event.id])

        response = self.get({"q": "Lunar", "ordering": "-live"})
        page_ids = [page.id for page in response.context["pages"]]
        self.assertEqual(page_ids, [live_event.id, draft_event.id])

    def test_search_filter_content_type(self):
        # Correct content_type
        response = self.get({"content_type": "demosite.standardpage"})
        self.assertEqual(response.status_code, 200)

        # Incorrect content_type
        response = self.get({"content_type": "demosite.standardpage.error"})
        self.assertEqual(response.status_code, 404)

    def test_empty_search_renders_content_type_facets(self):
        root_page = Page.objects.get(id=2)
        event_index = EventIndex(
            title="ALL THE EVENTS",
            intro="It's just a nod to the canon",
        )
        root_page.add_child(instance=event_index)

        params = [{"q": ""}, {}]
        url = reverse("wagtailadmin_pages:search")
        for param in params:
            with self.subTest(param=param):
                response = self.get(param)
                self.assertEqual(response.status_code, 200)
                self.assertTemplateUsed(response, "wagtailadmin/pages/search.html")
                self.assertEqual(response.context["query_string"], "")

                self.assertContains(response, "Page types")
                self.assertContains(response, "All (3)")
                # The test fixture contains the root page and the welcome page
                # with the base page type
                self.assertContains(response, "Page (2)")

                self.assertContains(response, "ALL THE EVENTS")

                self.assertContains(response, "Event index (1)")
                self.assertContains(
                    response,
                    f"{url}?q=&amp;content_type=tests.eventindex",
                )

                # The type column should not contain a link to order by content type
                soup = self.get_soup(response.content)
                headings = soup.select("main table thead th")
                type_th = None
                for heading in headings:
                    if heading.text.strip() == "Type":
                        type_th = heading
                self.assertIsNotNone(type_th)
                self.assertIsNone(type_th.select_one("a"))

    def test_empty_search_with_content_type_filter(self):
        root_page = Page.objects.get(id=2)
        event_index = EventIndex(
            title="ALL THE EVENTS",
            intro="It's just a nod to the canon",
        )
        new_event = SingleEventPage(
            title="Lunar event",
            location="the moon",
            audience="public",
            cost="free",
            date_from="2001-01-01",
            latest_revision_created_at=local_datetime(2016, 1, 1),
        )
        root_page.add_child(instance=event_index)
        root_page.add_child(instance=new_event)

        params = [
            {"q": "", "content_type": "tests.singleeventpage"},
            {"content_type": "tests.singleeventpage"},
        ]
        url = reverse("wagtailadmin_pages:search")
        for param in params:
            with self.subTest(param=param):
                response = self.get(param)
                self.assertEqual(response.status_code, 200)
                self.assertTemplateUsed(response, "wagtailadmin/pages/search.html")
                self.assertEqual(response.context["query_string"], "")

                self.assertContains(response, "Page types")
                self.assertContains(response, "All (4)")
                # The test fixture contains the root page and the welcome page
                # with the base page type
                self.assertContains(response, "Page (2)")
                self.assertContains(response, "Single event page (1)")

                self.assertContains(response, "Lunar event")
                self.assertNotContains(response, "ALL THE EVENTS")

                self.assertContains(response, "Event index (1)")
                self.assertContains(
                    response,
                    f"{url}?q=&amp;content_type=tests.eventindex",
                )



# Generated by Django 4.2.6 on 2023-10-27 18:03

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailsearch", "0007_delete_editorspick"),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name="querydailyhits",
            unique_together=None,
        ),
        migrations.RemoveField(
            model_name="querydailyhits",
            name="query",
        ),
        migrations.DeleteModel(
            name="Query",
        ),
        migrations.DeleteModel(
            name="QueryDailyHits",
        ),
    ]



from django.contrib.auth.models import Group, Permission
from django.test import TestCase
from django.urls import reverse

from wagtail.documents import get_document_model
from wagtail.models import Collection
from wagtail.test.utils import WagtailTestUtils

Document = get_document_model()


class TestBulkAddDocumentsToCollection(WagtailTestUtils, TestCase):
    def setUp(self):
        self.user = self.login()
        self.root_collection = Collection.get_first_root_node()
        self.dest_collection = self.root_collection.add_child(name="Destination")
        self.documents = [
            Document.objects.create(title=f"Test document - {i}") for i in range(1, 6)
        ]
        self.url = (
            reverse(
                "wagtail_bulk_action",
                args=(
                    "wagtaildocs",
                    "document",
                    "add_to_collection",
                ),
            )
            + "?"
        )
        for document in self.documents:
            self.url += f"id={document.id}&"
        self.post_data = {"collection": str(self.dest_collection.id)}

    def test_add_to_collection_with_limited_permissions(self):
        self.user.is_superuser = False
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            )
        )
        self.user.save()

        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)

        html = response.content.decode()
        self.assertInHTML(
            "<p>You don't have permission to add these documents to a collection</p>",
            html,
        )

        for document in self.documents:
            self.assertInHTML(f"<li>{document.title}</li>", html)

        self.client.post(self.url, self.post_data)

        # Documents should not be moved to new collection
        for document in self.documents:
            self.assertEqual(
                Document.objects.get(id=document.id).collection_id,
                self.root_collection.id,
            )

    def test_simple(self):
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(
            response, "wagtaildocs/bulk_actions/confirm_bulk_add_to_collection.html"
        )

    def test_add_to_collection(self):
        # Make post request
        response = self.client.post(self.url, self.post_data)

        # User should be redirected back to the index
        self.assertEqual(response.status_code, 302)

        # Documents should be moved to new collection
        for document in self.documents:
            self.assertEqual(
                Document.objects.get(id=document.id).collection_id,
                self.dest_collection.id,
            )


class TestBulkAddAllDocumentsToCollection(WagtailTestUtils, TestCase):
    def setUp(self):
        self.user = self.login()
        self.root_collection = Collection.get_first_root_node()
        self.dest_collection = self.root_collection.add_child(name="Destination")
        self.documents = [
            Document.objects.create(title=f"Test document - {i}") for i in range(1, 6)
        ]
        self.url = (
            reverse(
                "wagtail_bulk_action",
                args=(
                    "wagtaildocs",
                    "document",
                    "add_to_collection",
                ),
            )
            + "?id=all"
        )
        self.post_data = {"collection": str(self.dest_collection.id)}

    def test_add_all_to_collection_with_limited_permissions(self):
        # Create a group with document permissions on the source collection only
        self.other_collection = self.root_collection.add_child(name="Other")

        group = Group.objects.create(name="Other Collection Editors")
        self.other_collection.group_permissions.create(
            group=group, permission=Permission.objects.get(codename="change_document")
        )

        self.user.is_superuser = False
        self.user.groups.add(group)
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            )
        )
        self.user.save()

        # The user needs to have access to at least one document in order to access the bulk action page
        other_document = Document.objects.create(
            title="Test document - other", collection=self.other_collection
        )

        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)

        html = response.content.decode()

        self.assertInHTML(other_document.title, html)
        for document in self.documents:
            self.assertNotInHTML(f"<li>{document.title}</li>", html)

        self.client.post(self.url, self.post_data)

        # Documents should not be moved to new collection
        for document in self.documents:
            self.assertEqual(
                Document.objects.get(id=document.id).collection_id,
                self.root_collection.id,
            )

    def test_simple(self):
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(
            response, "wagtaildocs/bulk_actions/confirm_bulk_add_to_collection.html"
        )

    def test_add_all_to_collection(self):
        # Make post request
        response = self.client.post(self.url, self.post_data)

        # User should be redirected back to the index
        self.assertEqual(response.status_code, 302)

        # Documents should be moved to new collection
        for document in self.documents:
            self.assertEqual(
                Document.objects.get(id=document.id).collection_id,
                self.dest_collection.id,
            )



# Generated by Django 5.0.7 on 2024-08-06 07:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("tests", "0042_alter_customdocument_file_size_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="customimage",
            name="description",
            field=models.CharField(
                blank=True, default="", max_length=255, verbose_name="description"
            ),
        ),
        migrations.AddField(
            model_name="customimagefilepath",
            name="description",
            field=models.CharField(
                blank=True, default="", max_length=255, verbose_name="description"
            ),
        ),
        migrations.AddField(
            model_name="customimagewithauthor",
            name="description",
            field=models.CharField(
                blank=True, default="", max_length=255, verbose_name="description"
            ),
        ),
        migrations.AddField(
            model_name="customrestaurantimage",
            name="description",
            field=models.CharField(
                blank=True, default="", max_length=255, verbose_name="description"
            ),
        ),
    ]



import datetime

from django.conf import settings
from django.contrib.auth.models import Permission
from django.test import TestCase
from django.urls import reverse
from django.utils import timezone
from django.utils.formats import localize
from freezegun import freeze_time

from wagtail.admin.utils import get_user_display_name
from wagtail.models import Page
from wagtail.test.utils import WagtailTestUtils
from wagtail.test.utils.template_tests import AdminTemplateTestUtils


class TestWorkflowHistoryDetail(AdminTemplateTestUtils, WagtailTestUtils, TestCase):
    fixtures = ["test.json"]
    base_breadcrumb_items = []

    def setUp(self):
        self.user = self.create_test_user()
        self.login(self.user)

        self.christmas_event = Page.objects.get(
            url_path="/home/events/christmas/"
        ).specific

        self.site_root = Page.objects.specific().get(id=2)
        self.events_page = self.christmas_event.get_parent().specific

        workflow = self.christmas_event.get_workflow()

        self.timestamps = [
            datetime.datetime(2020, 1, 1, 10, 0, 0),
            datetime.datetime(2020, 1, 1, 11, 0, 0),
            datetime.datetime(2020, 1, 2, 12, 0, 0),
            datetime.datetime(2020, 1, 3, 13, 0, 0),
            datetime.datetime(2020, 1, 4, 14, 0, 0),
        ]

        if settings.USE_TZ:
            self.timestamps[:] = [
                timezone.make_aware(timestamp, timezone=datetime.timezone.utc)
                for timestamp in self.timestamps
            ]
            self.localized_timestamps = [
                localize(timezone.localtime(timestamp), "c")
                for timestamp in self.timestamps
            ]
        else:
            self.localized_timestamps = [
                localize(timestamp, "c") for timestamp in self.timestamps
            ]

        self.moderator = self.create_superuser("moderator")
        self.moderator_name = get_user_display_name(self.moderator)
        self.user_name = get_user_display_name(self.user)

        with freeze_time(self.timestamps[0]):
            self.christmas_event.save_revision()
        with freeze_time(self.timestamps[1]):
            self.workflow_state = workflow.start(self.christmas_event, self.user)

    def test_get_index(self):
        response = self.client.get(
            reverse(
                "wagtailadmin_pages:workflow_history", args=[self.christmas_event.id]
            )
        )
        self.assertEqual(response.status_code, 200)

        self.assertContains(
            response, reverse("wagtailadmin_pages:edit", args=[self.christmas_event.id])
        )
        self.assertContains(
            response,
            reverse(
                "wagtailadmin_pages:workflow_history_detail",
                args=[self.christmas_event.id, self.workflow_state.id],
            ),
        )

        # Should show the currently in progress workflow
        self.assertContains(response, "Moderators approval")
        self.assertContains(response, "In progress")
        self.assertContains(response, "test@email.com")

        items = [
            {
                "url": reverse("wagtailadmin_explore_root"),
                "label": "Root",
            },
            {
                "url": reverse("wagtailadmin_explore", args=(self.site_root.id,)),
                "label": self.site_root.get_admin_display_title(),
            },
            {
                "url": reverse("wagtailadmin_explore", args=(self.events_page.id,)),
                "label": self.events_page.get_admin_display_title(),
            },
            {
                "url": reverse("wagtailadmin_explore", args=(self.christmas_event.id,)),
                "label": self.christmas_event.get_admin_display_title(),
            },
            {
                "url": "",
                "label": "Workflow history",
                "sublabel": self.christmas_event.get_admin_display_title(),
            },
        ]
        self.assertBreadcrumbsItemsRendered(items, response.content)

    def test_get_index_with_bad_permissions(self):
        # Remove privileges from user
        self.user.is_superuser = False
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            )
        )
        self.user.save()

        response = self.client.get(
            reverse(
                "wagtailadmin_pages:workflow_history", args=[self.christmas_event.id]
            )
        )

        self.assertEqual(response.status_code, 302)

    def test_get_detail(self):
        task_state = self.workflow_state.current_task_state

        with freeze_time(self.timestamps[2]):
            task_state.task.on_action(
                task_state, user=self.moderator, action_name="reject"
            )
        self.workflow_state.refresh_from_db()

        with freeze_time(self.timestamps[3]):
            self.christmas_event.save_revision(user=self.user)
            self.workflow_state.resume(user=self.user)
        self.workflow_state.refresh_from_db()

        url = reverse(
            "wagtailadmin_pages:workflow_history_detail",
            args=[self.christmas_event.id, self.workflow_state.id],
        )
        self.client.get(url)

        with self.assertNumQueries(18):
            response = self.client.get(url)

        self.assertEqual(response.status_code, 200)

        self.assertContains(
            response, reverse("wagtailadmin_pages:edit", args=[self.christmas_event.id])
        )
        self.assertContains(
            response,
            reverse(
                "wagtailadmin_pages:workflow_history", args=[self.christmas_event.id]
            ),
        )

        soup = self.get_soup(response.content)

        self.assertIsNotNone(soup.select_one('[data-controller~="w-tabs"]'))
        self.assertIsNotNone(soup.select_one(".tab-content"))

        tasks = soup.select_one("#tab-tasks table")
        self.assertIsNotNone(tasks)
        cells = [
            [td.get_text(separator=" ", strip=True) for td in tr.select("td")]
            for tr in tasks.select("tr")
        ]

        self.assertEqual(
            cells,
            [
                # This is divided into different columns per task, so it makes
                # sense to start with the initial revision on the first cell and
                # then it should be rendered in ascending order.
                [
                    "Initial Revision",
                    f"Rejected by {self.moderator_name} at {self.localized_timestamps[2]}",
                ],
                [
                    f"Edited by {self.user_name} at {self.localized_timestamps[3]}",
                    "In progress",
                ],
            ],
        )

        timeline = soup.select_one("#tab-timeline table")
        self.assertIsNotNone(timeline)
        cells = [
            [td.get_text(separator=" ", strip=True) for td in tr.select("td")]
            for tr in timeline.select("tr")
        ]
        self.assertEqual(
            cells,
            [
                # The items are merged into a single column as a timeline, so it
                # should be rendered in reverse chronological order.
                [
                    self.localized_timestamps[3],
                    "Edited",
                ],
                [
                    self.localized_timestamps[2],
                    f"Moderators approval Rejected by {self.moderator_name}",
                ],
                [
                    self.localized_timestamps[1],
                    "Workflow started",
                ],
                [
                    self.localized_timestamps[0],
                    "Edited",
                ],
            ],
        )

        items = [
            {
                "url": reverse("wagtailadmin_explore_root"),
                "label": "Root",
            },
            {
                "url": reverse("wagtailadmin_explore", args=(self.site_root.id,)),
                "label": self.site_root.get_admin_display_title(),
            },
            {
                "url": reverse("wagtailadmin_explore", args=(self.events_page.id,)),
                "label": self.events_page.get_admin_display_title(),
            },
            {
                "url": reverse("wagtailadmin_explore", args=(self.christmas_event.id,)),
                "label": self.christmas_event.get_admin_display_title(),
            },
            {
                "url": reverse(
                    "wagtailadmin_pages:workflow_history",
                    args=(self.christmas_event.id,),
                ),
                "label": "Workflow history",
            },
            {
                "url": "",
                "label": "Workflow progress",
                "sublabel": self.christmas_event.get_admin_display_title(),
            },
        ]
        self.assertBreadcrumbsItemsRendered(items, response.content)

    def test_get_detail_completed(self):
        task_state = self.workflow_state.current_task_state

        with freeze_time(self.timestamps[2]):
            task_state.task.on_action(
                task_state, user=self.moderator, action_name="reject"
            )
        self.workflow_state.refresh_from_db()

        with freeze_time(self.timestamps[3]):
            self.christmas_event.save_revision(user=self.user)
            self.workflow_state.resume(user=self.user)
        self.workflow_state.refresh_from_db()

        with freeze_time(self.timestamps[4]):
            task_state = self.workflow_state.current_task_state
            task_state.task.on_action(
                task_state, user=self.moderator, action_name="approve"
            )
        self.workflow_state.refresh_from_db()

        url = reverse(
            "wagtailadmin_pages:workflow_history_detail",
            args=[self.christmas_event.id, self.workflow_state.id],
        )
        self.client.get(url)

        with self.assertNumQueries(19):
            response = self.client.get(url)

        self.assertEqual(response.status_code, 200)

        self.assertContains(
            response, reverse("wagtailadmin_pages:edit", args=[self.christmas_event.id])
        )
        self.assertContains(
            response,
            reverse(
                "wagtailadmin_pages:workflow_history", args=[self.christmas_event.id]
            ),
        )

        soup = self.get_soup(response.content)

        self.assertIsNotNone(soup.select_one('[data-controller~="w-tabs"]'))
        self.assertIsNotNone(soup.select_one(".tab-content"))

        tasks = soup.select_one("#tab-tasks table")
        self.assertIsNotNone(tasks)
        cells = [
            [td.get_text(separator=" ", strip=True) for td in tr.select("td")]
            for tr in tasks.select("tr")
        ]

        self.assertEqual(
            cells,
            [
                # This is divided into different columns per task, so it makes
                # sense to start with the initial revision on the first cell and
                # then it should be rendered in ascending order.
                [
                    "Initial Revision",
                    f"Rejected by {self.moderator_name} at {self.localized_timestamps[2]}",
                ],
                [
                    f"Edited by {self.user_name} at {self.localized_timestamps[3]}",
                    f"Approved by {self.moderator_name} at {self.localized_timestamps[4]}",
                ],
            ],
        )

        timeline = soup.select_one("#tab-timeline table")
        self.assertIsNotNone(timeline)
        cells = [
            [td.get_text(separator=" ", strip=True) for td in tr.select("td")]
            for tr in timeline.select("tr")
        ]
        self.assertEqual(
            cells,
            [
                # The items are merged into a single column as a timeline, so it
                # should be rendered in reverse chronological order.
                [
                    self.localized_timestamps[4],
                    "Workflow completed Approved",
                ],
                [
                    self.localized_timestamps[4],
                    f"Moderators approval Approved by {self.moderator_name}",
                ],
                [
                    self.localized_timestamps[3],
                    "Edited",
                ],
                [
                    self.localized_timestamps[2],
                    f"Moderators approval Rejected by {self.moderator_name}",
                ],
                [
                    self.localized_timestamps[1],
                    "Workflow started",
                ],
                [
                    self.localized_timestamps[0],
                    "Edited",
                ],
            ],
        )

    def test_get_detail_with_bad_permissions(self):
        # Remove privileges from user
        self.user.is_superuser = False
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            )
        )
        self.user.save()

        response = self.client.get(
            reverse(
                "wagtailadmin_pages:workflow_history_detail",
                args=[self.christmas_event.id, self.workflow_state.id],
            )
        )

        self.assertEqual(response.status_code, 302)



from django.apps import apps
from django.contrib.staticfiles.urls import staticfiles_urlpatterns
from django.http import HttpResponse
from django.urls import include, path

from wagtail import urls as wagtail_urls
from wagtail.admin import urls as wagtailadmin_urls
from wagtail.admin.views import home
from wagtail.api.v2.router import WagtailAPIRouter
from wagtail.api.v2.tests.test_pages import Test10411APIViewSet
from wagtail.api.v2.views import PagesAPIViewSet
from wagtail.contrib.redirects.api import RedirectsAPIViewSet
from wagtail.contrib.sitemaps import Sitemap
from wagtail.contrib.sitemaps import views as sitemaps_views
from wagtail.documents import urls as wagtaildocs_urls
from wagtail.documents.api.v2.views import DocumentsAPIViewSet
from wagtail.images import urls as wagtailimages_urls
from wagtail.images.api.v2.views import ImagesAPIViewSet
from wagtail.images.tests import urls as wagtailimages_test_urls
from wagtail.test.testapp import urls as testapp_urls
from wagtail.test.testapp.models import EventSitemap

api_router = WagtailAPIRouter("wagtailapi_v2")
api_router.register_endpoint("pages", PagesAPIViewSet)
api_router.register_endpoint("images", ImagesAPIViewSet)
api_router.register_endpoint("documents", DocumentsAPIViewSet)
api_router.register_endpoint("redirects", RedirectsAPIViewSet)
api_router.register_endpoint("issue_10411", Test10411APIViewSet)


urlpatterns = [
    path("admin/", include(wagtailadmin_urls)),
    path("documents/", include(wagtaildocs_urls)),
    path("testimages/", include(wagtailimages_test_urls)),
    path("images/", include(wagtailimages_urls)),
    path("api/main/", api_router.urls),
    path("sitemap.xml", sitemaps_views.sitemap),
    path(
        "sitemap-index.xml",
        sitemaps_views.index,
        {
            "sitemaps": {"pages": Sitemap, "events": EventSitemap(request=None)},
            "sitemap_url_name": "sitemap",
        },
    ),
    path("sitemap-<str:section>.xml", sitemaps_views.sitemap, name="sitemap"),
    path("testapp/", include(testapp_urls)),
    path("fallback/", lambda request: HttpResponse("ok"), name="fallback"),
]

if apps.is_installed("pattern_library"):
    urlpatterns += [
        path(
            "pattern-library/api/v1/sprite",
            home.sprite,
            name="pattern_library_sprite",
        ),
        path("pattern-library/", include("pattern_library.urls")),
    ]

urlpatterns += staticfiles_urlpatterns()

urlpatterns += [
    # For anything not caught by a more specific rule above, hand over to
    # Wagtail's serving mechanism
    path("", include(wagtail_urls)),
]



from django.utils.html import escape
from django.utils.translation import gettext_lazy as _

from wagtail.utils.apps import get_app_submodules

from .shortcuts import get_rendition_or_not_found


class Format:
    def __init__(self, name, label, classname, filter_spec):
        self.name = name
        self.label = label
        self.classname = classname
        self.filter_spec = filter_spec

    def __str__(self):
        return (
            f'"{self.name}", "{self.label}", "{self.classname}", "{self.filter_spec}"'
        )

    def __repr__(self):
        return f"Format({self})"

    def editor_attributes(self, image, alt_text):
        """
        Return additional attributes to go on the HTML element
        when outputting this image within a rich text editor field
        """
        return {
            "data-embedtype": "image",
            "data-id": image.id,
            "data-format": self.name,
            "data-alt": escape(alt_text),
        }

    def image_to_editor_html(self, image, alt_text):
        return self.image_to_html(
            image, alt_text, self.editor_attributes(image, alt_text)
        )

    def image_to_html(self, image, alt_text, extra_attributes=None):
        if extra_attributes is None:
            extra_attributes = {}
        rendition = get_rendition_or_not_found(image, self.filter_spec)

        extra_attributes["alt"] = escape(alt_text)
        if self.classname:
            extra_attributes["class"] = "%s" % escape(self.classname)

        return rendition.img_tag(extra_attributes)


FORMATS = []
FORMATS_BY_NAME = {}


def register_image_format(format):
    if format.name in FORMATS_BY_NAME:
        raise KeyError("Image format '%s' is already registered" % format.name)
    FORMATS_BY_NAME[format.name] = format
    FORMATS.append(format)


def unregister_image_format(format_name):
    global FORMATS
    # handle being passed a format object rather than a format name string
    try:
        format_name = format_name.name
    except AttributeError:
        pass

    try:
        del FORMATS_BY_NAME[format_name]
        FORMATS = [fmt for fmt in FORMATS if fmt.name != format_name]
    except KeyError:
        raise KeyError("Image format '%s' is not registered" % format_name)


def get_image_formats():
    search_for_image_formats()
    return FORMATS


def get_image_format(name):
    search_for_image_formats()
    return FORMATS_BY_NAME[name]


_searched_for_image_formats = False


def search_for_image_formats():
    global _searched_for_image_formats
    if not _searched_for_image_formats:
        list(get_app_submodules("image_formats"))
        _searched_for_image_formats = True


# Define default image formats
register_image_format(
    Format("fullwidth", _("Full width"), "richtext-image full-width", "width-800")
)
register_image_format(
    Format("left", _("Left-aligned"), "richtext-image left", "width-500")
)
register_image_format(
    Format("right", _("Right-aligned"), "richtext-image right", "width-500")
)



# Generated by Django 4.0.4 on 2022-09-09 14:52

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("tests", "0008_modelwithstringtypeprimarykey"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="eventpage",
            options={
                "permissions": [
                    ("custom_see_panel_setting", "Can see the panel."),
                    ("other_custom_see_panel_setting", "Can see the panel."),
                ]
            },
        ),
    ]



from django.test import TestCase, override_settings

from wagtail.users.utils import get_gravatar_url


class TestGravatar(TestCase):
    def test_gravatar_default(self):
        """Test with the default settings"""
        self.assertEqual(
            get_gravatar_url("something@example.com"),
            "//www.gravatar.com/avatar/76ebd6fecabc982c205dd056e8f0415a?d=mp&s=100",
        )

    def test_gravatar_custom_size(self):
        """Test with a custom size (note that the size will be doubled)"""
        self.assertEqual(
            get_gravatar_url("something@example.com", size=100),
            "//www.gravatar.com/avatar/76ebd6fecabc982c205dd056e8f0415a?d=mp&s=200",
        )

    @override_settings(
        WAGTAIL_GRAVATAR_PROVIDER_URL="https://robohash.org/avatar?d=robohash&s=200"
    )
    def test_gravatar_params_that_overlap(self):
        """
        Test with params that overlap with default s (size) and d (default_image)
        Also test the `s` is not overridden by the provider URL's query parameters.
        """
        self.assertEqual(
            get_gravatar_url("something@example.com", size=80),
            "https://robohash.org/avatar/76ebd6fecabc982c205dd056e8f0415a?d=robohash&s=160",
        )

    @override_settings(WAGTAIL_GRAVATAR_PROVIDER_URL="https://robohash.org/avatar?f=y")
    def test_gravatar_params_that_dont_overlap(self):
        """Test with params that don't default `s (size)` and `d (default_image)`"""
        self.assertEqual(
            get_gravatar_url("something@example.com"),
            "https://robohash.org/avatar/76ebd6fecabc982c205dd056e8f0415a?d=mp&f=y&s=100",
        )

    @override_settings(
        WAGTAIL_GRAVATAR_PROVIDER_URL="https://robohash.org/avatar?d=robohash&f=y"
    )
    def test_gravatar_query_params_override_default_params(self):
        """Test that query parameters of `WAGTAIL_GRAVATAR_PROVIDER_URL` override default_params"""
        self.assertEqual(
            get_gravatar_url(
                "something@example.com", default_params={"d": "monsterid"}
            ),
            "https://robohash.org/avatar/76ebd6fecabc982c205dd056e8f0415a?d=robohash&f=y&s=100",
        )

    @override_settings(WAGTAIL_GRAVATAR_PROVIDER_URL="https://robohash.org/avatar/")
    def test_gravatar_trailing_slash(self):
        """Test with a trailing slash in the URL"""
        self.assertEqual(
            get_gravatar_url("something@example.com"),
            "https://robohash.org/avatar/76ebd6fecabc982c205dd056e8f0415a?d=mp&s=100",
        )

    @override_settings(WAGTAIL_GRAVATAR_PROVIDER_URL="https://robohash.org/avatar")
    def test_gravatar_no_trailing_slash(self):
        """Test with no trailing slash in the URL"""
        self.assertEqual(
            get_gravatar_url("something@example.com"),
            "https://robohash.org/avatar/76ebd6fecabc982c205dd056e8f0415a?d=mp&s=100",
        )

    @override_settings(WAGTAIL_GRAVATAR_PROVIDER_URL="https://robohash.org/avatar?")
    def test_gravatar_trailing_question_mark(self):
        """Test with a trailing question mark in the URL"""
        self.assertEqual(
            get_gravatar_url("something@example.com"),
            "https://robohash.org/avatar/76ebd6fecabc982c205dd056e8f0415a?d=mp&s=100",
        )



# Generated by Django 5.1.4 on 2025-01-19 21:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("tests", "0045_alter_streampage_body"),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name="customrendition",
            unique_together=set(),
        ),
        migrations.AddConstraint(
            model_name="customrendition",
            constraint=models.UniqueConstraint(
                fields=("image", "filter_spec", "focal_point_key"),
                name="unique_rendition",
            ),
        ),
    ]



from django import forms
from django.conf import settings
from django.utils.translation import gettext as _
from django.utils.translation import ngettext

from wagtail.admin import widgets
from wagtail.models import Page, PageViewRestriction

from .models import WagtailAdminModelForm
from .view_restrictions import BaseViewRestrictionForm


class CopyForm(forms.Form):
    def __init__(self, *args, **kwargs):
        # CopyPage must be passed a 'page' kwarg indicating the page to be copied
        self.page = kwargs.pop("page")
        self.user = kwargs.pop("user", None)
        can_publish = kwargs.pop("can_publish")
        super().__init__(*args, **kwargs)
        self.fields["new_title"] = forms.CharField(
            initial=self.page.title, label=_("New title")
        )
        allow_unicode = getattr(settings, "WAGTAIL_ALLOW_UNICODE_SLUGS", True)
        self.fields["new_slug"] = forms.SlugField(
            initial=self.page.slug,
            label=_("New slug"),
            allow_unicode=allow_unicode,
            widget=widgets.SlugInput(locale=self.page.locale),
        )
        self.fields["new_parent_page"] = forms.ModelChoiceField(
            initial=self.page.get_parent(),
            queryset=Page.objects.all(),
            widget=widgets.AdminPageChooser(
                target_models=self.page.specific_class.allowed_parent_page_models(),
                can_choose_root=True,
                user_perms="copy_to",
            ),
            label=_("New parent page"),
            help_text=_("This copy will be a child of this given parent page."),
        )
        pages_to_copy = self.page.get_descendants(inclusive=True)
        subpage_count = pages_to_copy.count() - 1
        if subpage_count > 0:
            self.fields["copy_subpages"] = forms.BooleanField(
                required=False,
                initial=True,
                label=_("Copy subpages"),
                help_text=ngettext(
                    "This will copy %(count)s subpage.",
                    "This will copy %(count)s subpages.",
                    subpage_count,
                )
                % {"count": subpage_count},
            )

        if can_publish:
            pages_to_publish_count = pages_to_copy.live().count()
            if pages_to_publish_count > 0:
                # In the specific case that there are no subpages, customise the field label and help text
                if subpage_count == 0:
                    label = _("Publish copied page")
                    help_text = _(
                        "This page is live. Would you like to publish its copy as well?"
                    )
                else:
                    label = _("Publish copies")
                    help_text = ngettext(
                        "%(count)s of the pages being copied is live. Would you like to publish its copy?",
                        "%(count)s of the pages being copied are live. Would you like to publish their copies?",
                        pages_to_publish_count,
                    ) % {"count": pages_to_publish_count}

                self.fields["publish_copies"] = forms.BooleanField(
                    required=False, initial=False, label=label, help_text=help_text
                )

            # Note that only users who can publish in the new parent page can create an alias.
            # This is because alias pages must always match their original page's state.
            self.fields["alias"] = forms.BooleanField(
                required=False,
                initial=False,
                label=_("Alias"),
                help_text=_("Keep the new pages updated with future changes"),
            )

    def clean(self):
        cleaned_data = super().clean()

        # Make sure the slug isn't already in use
        slug = cleaned_data.get("new_slug")

        # New parent page given in form or parent of source, if parent_page is empty
        parent_page = cleaned_data.get("new_parent_page") or self.page.get_parent()

        # check if user is allowed to create a page at given location.
        if not parent_page.permissions_for_user(self.user).can_add_subpage():
            self._errors["new_parent_page"] = self.error_class(
                [
                    _('You do not have permission to copy to page "%(page_title)s"')
                    % {
                        "page_title": parent_page.specific_deferred.get_admin_display_title()
                    }
                ]
            )

        # Count the pages with the same slug within the context of our copy's parent page
        if slug and parent_page.get_children().filter(slug=slug).count():
            self._errors["new_slug"] = self.error_class(
                [
                    _(
                        'This slug is already in use within the context of its parent page "%(parent_page_title)s"'
                    )
                    % {"parent_page_title": parent_page}
                ]
            )
            # The slug is no longer valid, hence remove it from cleaned_data
            del cleaned_data["new_slug"]

        # Don't allow recursive copies into self
        if cleaned_data.get("copy_subpages") and (
            self.page == parent_page or parent_page.is_descendant_of(self.page)
        ):
            self._errors["new_parent_page"] = self.error_class(
                [_("You cannot copy a page into itself when copying subpages")]
            )

        return cleaned_data


class PageViewRestrictionForm(BaseViewRestrictionForm):
    def __init__(self, *args, **kwargs):
        # get the list of private page options from the page
        private_page_options = kwargs.pop("private_page_options", [])

        super().__init__(*args, **kwargs)

        if not getattr(settings, "WAGTAIL_PRIVATE_PAGE_OPTIONS", {}).get(
            "SHARED_PASSWORD", True
        ):
            self.fields["restriction_type"].choices = [
                choice
                for choice in PageViewRestriction.RESTRICTION_CHOICES
                if choice[0] != PageViewRestriction.PASSWORD
            ]
            del self.fields["password"]
        # Remove the fields that are not allowed for the page
        self.fields["restriction_type"].choices = [
            choice
            for choice in self.fields["restriction_type"].choices
            if choice[0] in private_page_options
            or choice[0] == PageViewRestriction.NONE
        ]

    class Meta:
        model = PageViewRestriction
        fields = ("restriction_type", "password", "groups")


class WagtailAdminPageForm(WagtailAdminModelForm):
    comment_notifications = forms.BooleanField(
        widget=forms.CheckboxInput(), required=False
    )

    def __init__(
        self,
        data=None,
        files=None,
        parent_page=None,
        subscription=None,
        *args,
        **kwargs,
    ):
        self.subscription = subscription

        initial = kwargs.pop("initial", {})
        if self.subscription:
            initial["comment_notifications"] = subscription.comment_notifications

        super().__init__(data, files, *args, initial=initial, **kwargs)

        self.parent_page = parent_page

        if not self.show_comments_toggle:
            del self.fields["comment_notifications"]

    @property
    def show_comments_toggle(self):
        return "comments" in self.__class__.formsets

    def save(self, commit=True):
        # Save comment notifications updates to PageSubscription
        if self.show_comments_toggle and self.subscription:
            self.subscription.comment_notifications = self.cleaned_data[
                "comment_notifications"
            ]
            if commit:
                self.subscription.save()

        return super().save(commit=commit)

    def is_valid(self):
        comments = self.formsets.get("comments")
        # Remove the comments formset if the management form is invalid
        if comments and not comments.management_form.is_valid():
            del self.formsets["comments"]
        return super().is_valid()

    def clean(self):
        cleaned_data = super().clean()
        if "slug" in self.cleaned_data:
            page_slug = cleaned_data["slug"]
            if not Page._slug_is_available(page_slug, self.parent_page, self.instance):
                self.add_error(
                    "slug",
                    forms.ValidationError(
                        _(
                            "The slug '%(page_slug)s' is already in use within the parent page"
                        )
                        % {"page_slug": page_slug}
                    ),
                )

        return cleaned_data

    @property
    def media(self):
        media = super().media
        if self.show_comments_toggle:
            media += forms.Media(
                js=["wagtailadmin/js/comments.js"],
            )
        return media


class MoveForm(forms.Form):
    def __init__(self, *args, **kwargs):
        self.page_to_move = kwargs.pop("page_to_move")
        self.target_parent_models = kwargs.pop("target_parent_models")

        super().__init__(*args, **kwargs)

        self.fields["new_parent_page"] = forms.ModelChoiceField(
            initial=self.page_to_move.get_parent(),
            queryset=Page.objects.all(),
            widget=widgets.AdminPageMoveChooser(
                can_choose_root=True,
                user_perms="move_to",
                target_models=self.target_parent_models,
                pages_to_move=[self.page_to_move.pk],
            ),
            label=_("New parent page"),
            help_text=_("Select a new parent for this page."),
        )


class ParentChooserForm(forms.Form):
    def __init__(self, child_page_type, user, *args, **kwargs):
        self.child_page_type = child_page_type
        self.user = user
        super().__init__(*args, **kwargs)
        self.fields["parent_page"] = forms.ModelChoiceField(
            queryset=Page.objects.all(),
            widget=widgets.AdminPageChooser(
                target_models=self.child_page_type.allowed_parent_page_models(),
                can_choose_root=True,
                user_perms="add_subpage",
            ),
            label=_("Parent page"),
            help_text=_("The new page will be a child of this given parent page."),
        )

    def clean_parent_page(self):
        parent_page = self.cleaned_data["parent_page"].specific_deferred
        if not parent_page.permissions_for_user(self.user).can_add_subpage():
            raise forms.ValidationError(
                _('You do not have permission to create a page under "%(page_title)s".')
                % {"page_title": parent_page.get_admin_display_title()}
            )
        if not self.child_page_type.can_create_at(parent_page):
            raise forms.ValidationError(
                _(
                    'You cannot create a page of type "%(page_type)s" under "%(page_title)s".'
                )
                % {
                    "page_type": self.child_page_type.get_verbose_name(),
                    "page_title": parent_page.get_admin_display_title(),
                }
            )
        return parent_page



from django import forms


class BaseFormSetMixin:
    """
    A mixin for formsets that adds the necessary attributes for the w-formset controller.
    So that JavaScript behavior can be added to the formset for dynamic addition and deletion of child forms.
    See client/src/controllers/FormsetController.ts
    """

    deletion_widget = forms.HiddenInput(attrs={"data-w-formset-target": "deleteInput"})

    @property
    def attrs(self):
        return {
            "data-controller": "w-formset",
            "data-w-formset-deleted-class": "w-transition-opacity w-duration-300 w-ease-out w-opacity-0",
        }

    @property
    def management_form(self):
        form = super().management_form

        for field in form:
            if field.name.endswith(forms.formsets.TOTAL_FORM_COUNT):
                field.field.widget.attrs["data-w-formset-target"] = "totalFormsInput"
            if field.name.endswith(forms.formsets.MIN_NUM_FORM_COUNT):
                field.field.widget.attrs["data-w-formset-target"] = "minFormsInput"
            if field.name.endswith(forms.formsets.MAX_NUM_FORM_COUNT):
                field.field.widget.attrs["data-w-formset-target"] = "maxFormsInput"

        return form



# Generated by Django 5.0.6 on 2024-09-23 15:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailusers", "0013_userprofile_density"),
    ]

    operations = [
        migrations.AddField(
            model_name="userprofile",
            name="contrast",
            field=models.CharField(
                choices=[("system", "System default"), ("more_contrast", "More contrast")],
                default="system",
                max_length=40,
                verbose_name="contrast",
            ),
        ),
    ]



from django.contrib import messages
from django.contrib.admin.utils import unquote
from django.core.exceptions import PermissionDenied
from django.db import transaction
from django.http import Http404
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy
from django.views.generic import TemplateView
from django.views.generic.detail import SingleObjectMixin

from wagtail.actions.copy_for_translation import CopyPageForTranslationAction
from wagtail.models import DraftStateMixin, Page, TranslatableMixin
from wagtail.snippets.views.snippets import get_snippet_model_from_url_params

from .forms import SubmitTranslationForm


class SubmitTranslationView(SingleObjectMixin, TemplateView):
    template_name = "simple_translation/admin/submit_translation.html"
    title = gettext_lazy("Translate")

    def get_title(self):
        return self.title

    def get_subtitle(self):
        return str(self.object)

    def get_form(self):
        if self.request.method == "POST":
            return SubmitTranslationForm(self.object, self.request.POST)
        return SubmitTranslationForm(self.object)

    def get_success_url(self, translated_object=None):
        raise NotImplementedError

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(
            {
                "form": self.get_form(),
            }
        )
        return context

    def post(self, request, **kwargs):  # pragma: no mccabe
        form = self.get_form()

        if form.is_valid():
            include_subtree = form.cleaned_data["include_subtree"]
            user = request.user

            with transaction.atomic():
                for locale in form.cleaned_data["locales"]:
                    if isinstance(self.object, Page):
                        action = CopyPageForTranslationAction(
                            page=self.object,
                            locale=locale,
                            include_subtree=include_subtree,
                            user=user,
                        )
                        action.execute(skip_permission_checks=True)

                    else:
                        self.object.copy_for_translation(
                            locale
                        ).save()  # pragma: no cover

                single_translated_object = None
                if len(form.cleaned_data["locales"]) == 1:
                    locales = form.cleaned_data["locales"][0].get_display_name()
                    single_translated_object = self.object.get_translation(
                        form.cleaned_data["locales"][0]
                    )
                else:
                    # Translators: always plural
                    locales = _("%(locales_count)s locales") % {
                        "locales_count": len(form.cleaned_data["locales"])
                    }

                messages.success(self.request, self.get_success_message(locales))

                return redirect(self.get_success_url(single_translated_object))

        context = self.get_context_data(**kwargs)
        return self.render_to_response(context)

    def dispatch(self, request, *args, **kwargs):
        if not request.user.has_perms(["simple_translation.submit_translation"]):
            raise PermissionDenied

        self.object = self.get_object()
        return super().dispatch(request, *args, **kwargs)


class SubmitPageTranslationView(SubmitTranslationView):
    title = gettext_lazy("Translate page")

    def get_subtitle(self):
        return self.object.get_admin_display_title()

    def get_object(self):
        page = get_object_or_404(Page, id=self.kwargs["page_id"]).specific

        # Can't translate the root page
        if page.is_root():
            raise Http404

        return page

    def get_success_url(self, translated_page=None):
        if translated_page:
            # If the editor chose a single locale to translate to, redirect to
            # the newly translated page's edit view.
            return reverse("wagtailadmin_pages:edit", args=[translated_page.id])

        return reverse("wagtailadmin_explore", args=[self.get_object().get_parent().id])

    def get_success_message(self, locales):
        return _(
            "The page '%(page_title)s' was successfully created in %(locales)s"
        ) % {"page_title": self.object.get_admin_display_title(), "locales": locales}


class SubmitSnippetTranslationView(SubmitTranslationView):
    def get_title(self):
        return _("Translate %(model_name)s") % {
            "model_name": self.object._meta.verbose_name
        }

    def get_object(self):
        model = get_snippet_model_from_url_params(
            self.kwargs["app_label"], self.kwargs["model_name"]
        )

        if not issubclass(model, TranslatableMixin):
            raise Http404

        object = get_object_or_404(model, pk=unquote(str(self.kwargs["pk"])))
        if isinstance(object, DraftStateMixin):
            object = object.get_latest_revision_as_object()

        return object

    def get_success_url(self, translated_snippet=None):
        pk = self.kwargs["pk"]

        if translated_snippet:
            # If the editor chose a single locale to translate to, redirect to
            # the newly translated snippet's edit view.
            pk = translated_snippet.pk

        return reverse(self.object.snippet_viewset.get_url_name("edit"), args=[pk])

    def get_success_message(self, locales):
        return _("Successfully created %(locales)s for %(model_name)s '%(object)s'") % {
            "model_name": self.object._meta.verbose_name,
            "object": str(self.object),
            "locales": locales,
        }



from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext

from wagtail.admin.views.pages.bulk_actions.page_bulk_action import PageBulkAction


class UnpublishBulkAction(PageBulkAction):
    display_name = _("Unpublish")
    action_type = "unpublish"
    aria_label = _("Unpublish selected pages")
    template_name = "wagtailadmin/pages/bulk_actions/confirm_bulk_unpublish.html"
    action_priority = 50

    def check_perm(self, page):
        return page.permissions_for_user(self.request.user).can_unpublish()

    def object_context(self, page):
        return {
            **super().object_context(page),
            "live_descendant_count": page.get_descendants().live().count(),
        }

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["has_live_descendants"] = any(
            item["live_descendant_count"] > 0 for item in context["items"]
        )
        return context

    def get_execution_context(self):
        return {
            **super().get_execution_context(),
            "permission_checker": self.check_perm,
            "include_descendants": self.cleaned_form.cleaned_data[
                "include_descendants"
            ],
        }

    @classmethod
    def execute_action(
        cls,
        objects,
        include_descendants=False,
        user=None,
        permission_checker=None,
        **kwargs,
    ):
        num_parent_objects, num_child_objects = 0, 0
        for page in objects:
            page.unpublish(user=user)
            num_parent_objects += 1

            if include_descendants:
                for live_descendant_page in (
                    page.get_descendants()
                    .live()
                    .defer_streamfields()
                    .specific()
                    .iterator()
                ):
                    if user is None or permission_checker(live_descendant_page):
                        live_descendant_page.unpublish()
                        num_child_objects += 1
        return num_parent_objects, num_child_objects

    def get_success_message(self, num_parent_objects, num_child_objects):
        include_descendants = self.cleaned_form.cleaned_data["include_descendants"]
        if include_descendants and num_child_objects > 0:
            # Translators: This forms a message such as "1 page and 3 child pages have been unpublished"
            return _("%(parent_pages)s and %(child_pages)s have been unpublished") % {
                "parent_pages": self.get_parent_page_text(num_parent_objects),
                "child_pages": self.get_child_page_text(num_child_objects),
            }
        else:
            return ngettext(
                "%(num_parent_objects)d page has been unpublished",
                "%(num_parent_objects)d pages have been unpublished",
                num_parent_objects,
            ) % {"num_parent_objects": num_parent_objects}



from django.contrib.contenttypes.models import ContentType
from django.db.models import DEFERRED
from django.utils.functional import cached_property


class SpecificMixin:
    """
    Mixin for models that support multi-table inheritance and provide a
    ``content_type`` field pointing to the specific model class, to provide
    methods and properties for retrieving the specific instance of the model.
    """

    def get_specific(self, deferred=False, copy_attrs=None, copy_attrs_exclude=None):
        """
        Return this object in its most specific subclassed form.

        By default, a database query is made to fetch all field values for the
        specific object. If you only require access to custom methods or other
        non-field attributes on the specific object, you can use
        ``deferred=True`` to avoid this query. However, any attempts to access
        specific field values from the returned object will trigger additional
        database queries.

        By default, references to all non-field attribute values are copied
        from current object to the returned one. This includes:

        * Values set by a queryset, for example: annotations, or values set as
          a result of using ``select_related()`` or ``prefetch_related()``.
        * Any ``cached_property`` values that have been evaluated.
        * Attributes set elsewhere in Python code.

        For fine-grained control over which non-field values are copied to the
        returned object, you can use ``copy_attrs`` to specify a complete list
        of attribute names to include. Alternatively, you can use
        ``copy_attrs_exclude`` to specify a list of attribute names to exclude.

        If called on an object that is already an instance of the most specific
        class, the object will be returned as is, and no database queries or
        other operations will be triggered.

        If the object was originally created using a model that has since
        been removed from the codebase, an instance of the base class will be
        returned (without any custom field values or other functionality
        present on the original class). Usually, deleting these objects is the
        best course of action, but there is currently no safe way for Wagtail
        to do that at migration time.
        """
        model_class = self.specific_class

        if model_class is None:
            # The codebase and database are out of sync (e.g. the model exists
            # on a different git branch and migrations were not applied or
            # reverted before switching branches). So, the best we can do is
            # return the page in it's current form.
            return self

        if isinstance(self, model_class):
            # self is already an instance of the most specific class.
            return self

        if deferred:
            # Generate a tuple of values in the order expected by __init__(),
            # with missing values substituted with DEFERRED ()
            values = tuple(
                getattr(self, f.attname, self.pk if f.primary_key else DEFERRED)
                for f in model_class._meta.concrete_fields
            )
            # Create object from known attribute values
            specific_obj = model_class(*values)
            specific_obj._state.adding = self._state.adding
        else:
            # Fetch object from database
            specific_obj = model_class._default_manager.get(id=self.id)

        # Copy non-field attribute values
        if copy_attrs is not None:
            for attr in (attr for attr in copy_attrs if attr in self.__dict__):
                setattr(specific_obj, attr, getattr(self, attr))
        else:
            exclude = copy_attrs_exclude or ()
            for k, v in ((k, v) for k, v in self.__dict__.items() if k not in exclude):
                # only set values that haven't already been set
                specific_obj.__dict__.setdefault(k, v)

        return specific_obj

    @cached_property
    def specific(self):
        """
        Returns this object in its most specific subclassed form with all field
        values fetched from the database. The result is cached in memory.
        """
        return self.get_specific()

    @cached_property
    def specific_deferred(self):
        """
        Returns this object in its most specific subclassed form without any
        additional field values being fetched from the database. The result
        is cached in memory.
        """
        return self.get_specific(deferred=True)

    @cached_property
    def specific_class(self):
        """
        Return the class that this object would be if instantiated in its
        most specific form.

        If the model class can no longer be found in the codebase, and the
        relevant ``ContentType`` has been removed by a database migration,
        the return value will be ``None``.

        If the model class can no longer be found in the codebase, but the
        relevant ``ContentType`` is still present in the database (usually a
        result of switching between git branches without running or reverting
        database migrations beforehand), the return value will be ``None``.
        """
        return self.cached_content_type.model_class()

    @property
    def cached_content_type(self):
        """
        Return this object's ``content_type`` value from the ``ContentType``
        model's cached manager, which will avoid a database query if the
        content type is already in memory.
        """
        return ContentType.objects.get_for_id(self.content_type_id)



from django.test import TestCase

from wagtail.models import Page, Site
from wagtail.test.testapp.models import EventIndex


class AllowedHttpMethodsTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        site = Site.objects.select_related("root_page").get(is_default_site=True)
        cls.page = Page(title="Page", slug="page")
        site.root_page.add_child(instance=cls.page)
        cls.event_index_page = EventIndex(title="Event index", slug="event-index")
        site.root_page.add_child(instance=cls.event_index_page)

    def test_options_request_for_default_page(self):
        response = self.client.options(self.page.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response["Allow"], "DELETE, GET, HEAD, OPTIONS, PATCH, POST, PUT"
        )

    def test_options_request_for_restricted_page(self):
        response = self.client.options(self.event_index_page.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response["Allow"], "GET, OPTIONS")

    def test_invalid_request_method_for_restricted_page(self):
        response = self.client.post(self.event_index_page.url)
        self.assertEqual(response.status_code, 405)
        self.assertEqual(response["Allow"], "GET, OPTIONS")



from io import StringIO

from django.contrib.admin.utils import quote
from django.contrib.auth.models import Permission
from django.core import management
from django.test import TestCase
from django.urls import reverse

from wagtail.models import Page, ReferenceIndex
from wagtail.test.testapp.models import (
    Advert,
    DraftStateModel,
    EventPage,
    GenericSnippetPage,
)
from wagtail.test.utils import WagtailTestUtils


class TestUsageCount(TestCase):
    fixtures = ["test.json"]

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        output = StringIO()
        management.call_command("rebuild_references_index", stdout=output)

    def test_snippet_usage_count(self):
        advert = Advert.objects.get(pk=1)
        self.assertEqual(ReferenceIndex.get_grouped_references_to(advert).count(), 2)


class TestUsedBy(TestCase):
    fixtures = ["test.json"]

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        output = StringIO()
        management.call_command("rebuild_references_index", stdout=output)

    def test_snippet_used_by(self):
        advert = Advert.objects.get(pk=1)
        usage = ReferenceIndex.get_grouped_references_to(advert)

        self.assertIsInstance(usage[0], tuple)
        self.assertIsInstance(usage[0][0], Page)
        self.assertIsInstance(usage[0][1], list)
        self.assertIsInstance(usage[0][1][0], ReferenceIndex)


class TestSnippetUsageView(WagtailTestUtils, TestCase):
    fixtures = ["test.json"]

    def setUp(self):
        self.user = self.login()

    def test_use_latest_draft_as_title(self):
        snippet = DraftStateModel.objects.create(text="Draft-enabled Foo, Published")
        snippet.save_revision().publish()
        snippet.text = "Draft-enabled Bar, In Draft"
        snippet.save_revision()

        response = self.client.get(
            reverse(
                "wagtailsnippets_tests_draftstatemodel:usage",
                args=[quote(snippet.pk)],
            )
        )

        soup = self.get_soup(response.content)
        sublabel = soup.select_one(".w-breadcrumbs__sublabel")
        # Should use the latest draft title in the breadcrumbs sublabel
        self.assertEqual(sublabel.get_text(strip=True), "Draft-enabled Bar, In Draft")

    def test_usage(self):
        with self.captureOnCommitCallbacks(execute=True):
            # resave so that usage count gets updated
            page = Page.objects.get(pk=2)
            page.save()

            gfk_page = GenericSnippetPage(
                title="Foobar Title",
                snippet_content_object=Advert.objects.get(pk=1),
            )
            page.add_child(instance=gfk_page)

        response = self.client.get(
            reverse(
                "wagtailsnippets_tests_advert:usage",
                args=["1"],
            )
        )
        self.assertContains(response, "Welcome to the Wagtail test site!")
        self.assertContains(response, "Foobar Title")
        self.assertContains(response, "<td>Generic snippet page</td>", html=True)
        self.assertContains(response, "Snippet content object")
        self.assertContains(response, "<th>Field</th>", html=True)
        self.assertNotContains(response, "<th>If you confirm deletion</th>", html=True)
        self.assertContains(response, "Snippet content object")
        self.assertContains(
            response,
            reverse("wagtailadmin_pages:edit", args=[gfk_page.id])
            + "#:w:contentpath=snippet_content_object",
        )

    def test_usage_without_edit_permission_on_snippet(self):
        # Create a user with basic admin backend access
        user = self.create_user(
            username="basicadmin", email="basicadmin@example.com", password="password"
        )
        admin_permission = Permission.objects.get(
            content_type__app_label="wagtailadmin", codename="access_admin"
        )
        user.user_permissions.add(admin_permission)
        self.login(username="basicadmin", password="password")

        response = self.client.get(
            reverse(
                "wagtailsnippets_tests_advert:usage",
                args=["1"],
            )
        )
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, reverse("wagtailadmin_home"))

    def test_usage_without_edit_permission_on_page(self):
        with self.captureOnCommitCallbacks(execute=True):
            # resave so that usage count gets updated
            page = Page.objects.get(pk=2)
            page.save()

        # Create a user with edit access to snippets but not pages
        user = self.create_user(
            username="basicadmin", email="basicadmin@example.com", password="password"
        )
        admin_permission = Permission.objects.get(
            content_type__app_label="wagtailadmin", codename="access_admin"
        )
        advert_permission = Permission.objects.get(
            content_type__app_label="tests", codename="change_advert"
        )
        user.user_permissions.add(admin_permission)
        user.user_permissions.add(advert_permission)
        self.login(username="basicadmin", password="password")

        response = self.client.get(
            reverse(
                "wagtailsnippets_tests_advert:usage",
                args=["1"],
            )
        )
        self.assertEqual(response.status_code, 200)
        self.assertNotContains(response, "Welcome to the Wagtail test site!")
        self.assertContains(response, "(Private page)")
        self.assertContains(response, "<td>Page</td>", html=True)
        self.assertContains(response, "<th>Field</th>", html=True)
        self.assertNotContains(response, "<th>If you confirm deletion</th>", html=True)
        self.assertContains(response, "<li>Advert</li>", html=True)

    def test_usage_with_describe_on_delete_cascade(self):
        with self.captureOnCommitCallbacks(execute=True):
            # resave so that usage count gets updated
            page = Page.objects.get(pk=2)
            page.save()

        response = self.client.get(
            reverse("wagtailsnippets_tests_advert:usage", args=["1"])
            + "?describe_on_delete=1"
        )
        self.assertContains(response, "Welcome to the Wagtail test site!")
        self.assertContains(response, "<td>Page</td>", html=True)
        self.assertNotContains(response, "<th>Field</th>", html=True)
        self.assertContains(response, "<th>If you confirm deletion</th>", html=True)
        self.assertContains(response, "Advert")
        self.assertContains(response, ": the advert placement will also be deleted")

    def test_usage_with_describe_on_delete_set_null(self):
        with self.captureOnCommitCallbacks(execute=True):
            # resave so that usage count gets updated
            page = EventPage.objects.first()
            page.save()

        self.assertEqual(page.feed_image.get_usage().count(), 1)

        response = self.client.get(
            reverse("wagtailimages:image_usage", args=[page.feed_image.id])
            + "?describe_on_delete=1"
        )
        self.assertContains(response, page.title)
        self.assertContains(response, "<td>Event page</td>", html=True)
        self.assertNotContains(response, "<th>Field</th>", html=True)
        self.assertContains(response, "<th>If you confirm deletion</th>", html=True)
        self.assertContains(response, "Feed image")
        self.assertContains(response, ": will unset the reference")

    def test_usage_with_describe_on_delete_gfk(self):
        with self.captureOnCommitCallbacks(execute=True):
            advert = Advert.objects.get(pk=1)

            gfk_page = GenericSnippetPage(
                title="Foobar Title",
                snippet_content_object=advert,
            )
            Page.objects.get(pk=1).add_child(instance=gfk_page)

        self.assertEqual(ReferenceIndex.get_grouped_references_to(advert).count(), 1)

        response = self.client.get(
            reverse("wagtailsnippets_tests_advert:usage", args=["1"])
            + "?describe_on_delete=1"
        )
        self.assertNotContains(response, "Welcome to the Wagtail test site!")
        self.assertContains(response, "Foobar Title")
        self.assertContains(response, "<td>Generic snippet page</td>", html=True)
        self.assertNotContains(response, "<th>Field</th>", html=True)
        self.assertContains(response, "<th>If you confirm deletion</th>", html=True)
        self.assertContains(response, "Snippet content object")
        self.assertContains(response, ": will unset the reference")



from wagtail.admin.mail import (
    GroupApprovalTaskStateSubmissionEmailNotifier,
    WorkflowStateApprovalEmailNotifier,
    WorkflowStateRejectionEmailNotifier,
    WorkflowStateSubmissionEmailNotifier,
)
from wagtail.models import TaskState, WorkflowState
from wagtail.signals import (
    task_submitted,
    workflow_approved,
    workflow_rejected,
    workflow_submitted,
)

task_submission_email_notifier = GroupApprovalTaskStateSubmissionEmailNotifier()
workflow_submission_email_notifier = WorkflowStateSubmissionEmailNotifier()
workflow_approval_email_notifier = WorkflowStateApprovalEmailNotifier()
workflow_rejection_email_notifier = WorkflowStateRejectionEmailNotifier()


def register_signal_handlers():
    task_submitted.connect(
        task_submission_email_notifier,
        sender=TaskState,
        dispatch_uid="group_approval_task_submitted_email_notification",
    )

    workflow_submitted.connect(
        workflow_submission_email_notifier,
        sender=WorkflowState,
        dispatch_uid="workflow_state_submitted_email_notification",
    )
    workflow_rejected.connect(
        workflow_rejection_email_notifier,
        sender=WorkflowState,
        dispatch_uid="workflow_state_rejected_email_notification",
    )
    workflow_approved.connect(
        workflow_approval_email_notifier,
        sender=WorkflowState,
        dispatch_uid="workflow_state_approved_email_notification",
    )



from unittest import mock

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.contrib.messages import constants as message_constants
from django.http import HttpRequest, HttpResponse
from django.test import TestCase
from django.urls import reverse

from wagtail.models import Page
from wagtail.signals import post_page_move, pre_page_move
from wagtail.test.testapp.models import BusinessSubIndex, SimplePage
from wagtail.test.utils import WagtailTestUtils


class TestPageMove(WagtailTestUtils, TestCase):
    fixtures = ["test.json"]

    @classmethod
    def setUpTestData(cls):
        # Find root page
        cls.root_page = Page.objects.get(id=2)

        # Create three sections
        cls.section_a = SimplePage(title="Section A", slug="section-a", content="hello")
        cls.root_page.add_child(instance=cls.section_a)

        cls.section_b = SimplePage(title="Section B", slug="section-b", content="hello")
        cls.root_page.add_child(instance=cls.section_b)

        cls.section_c = SimplePage(title="Section C", slug="section-c", content="hello")
        cls.root_page.add_child(instance=cls.section_c)

        # Add test page A into section A
        cls.test_page_a = SimplePage(
            title="Hello world!", slug="hello-world", content="hello"
        )
        cls.section_a.add_child(instance=cls.test_page_a)

        # Add test page B into section C
        cls.test_page_b = SimplePage(
            title="Hello world!", slug="hello-world", content="hello"
        )
        cls.section_c.add_child(instance=cls.test_page_b)

        # Add unpublished page to the root with a child page
        cls.unpublished_page = SimplePage(
            title="Unpublished", slug="unpublished", content="hello"
        )
        sub_page = SimplePage(title="Sub Page", slug="sub-page", content="child")
        cls.root_page.add_child(instance=cls.unpublished_page)
        cls.unpublished_page.add_child(instance=sub_page)

        # unpublish pages last (used to validate the edit only permission)
        cls.unpublished_page.unpublish()
        sub_page.unpublish()

    def setUp(self):
        # Login
        self.user = self.login()

    def test_page_move(self):
        response = self.client.get(
            reverse("wagtailadmin_pages:move", args=(self.test_page_a.id,))
        )
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, self.section_a.title)

    def test_page_move_bad_permissions(self):
        # Remove privileges from user
        self.user.is_superuser = False
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            )
        )
        self.user.save()

        # Get move page
        response = self.client.get(
            reverse("wagtailadmin_pages:move", args=(self.test_page_a.id,))
        )

        # Check that the user received a 302 redirected response
        self.assertEqual(response.status_code, 302)

    def test_user_without_bulk_delete_permission_can_move(self):
        # to verify that a user without bulk delete permission is able to move a page with a child page

        self.client.logout()
        user = get_user_model().objects.get(email="siteeditor@example.com")
        self.login(user)

        # ensure the bulk_delete is not applicable to this user
        can_bulk_delete = self.test_page_b.permissions_for_user(user).can_delete()
        self.assertFalse(can_bulk_delete)

        response = self.client.get(
            reverse("wagtailadmin_pages:move", args=(self.unpublished_page.id,))
        )

        self.assertEqual(response.status_code, 200)

    def test_page_move_confirm(self):
        response = self.client.get(
            reverse(
                "wagtailadmin_pages:move_confirm",
                args=(self.test_page_a.id, self.section_b.id),
            )
        )
        self.assertEqual(response.status_code, 200)

        response = self.client.get(
            reverse(
                "wagtailadmin_pages:move_confirm",
                args=(self.test_page_b.id, self.section_a.id),
            )
        )
        # Duplicate slugs triggers a redirect with an error message.
        self.assertEqual(response.status_code, 302)

        response = self.client.get(reverse("wagtailadmin_home"))
        messages = list(response.context["messages"])
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0].level, message_constants.ERROR)
        # Slug should be in error message.
        self.assertIn(self.test_page_b.slug, messages[0].message)

    def test_move_triggers_signals(self):
        # Connect a mock signal handler to pre_page_move and post_page_move signals
        pre_moved_handler = mock.MagicMock()
        post_moved_handler = mock.MagicMock()

        pre_page_move.connect(pre_moved_handler)
        post_page_move.connect(post_moved_handler)

        # Post to view to move page
        try:
            self.client.post(
                reverse(
                    "wagtailadmin_pages:move_confirm",
                    args=(self.test_page_a.id, self.section_b.id),
                )
            )
        finally:
            # Disconnect mock handler to prevent cross-test pollution
            pre_page_move.disconnect(pre_moved_handler)
            post_page_move.disconnect(post_moved_handler)

        # parent_page_before returns the non-specific page type, and that's OK
        nonspecific_section_a = Page.objects.get(pk=self.section_a.pk)

        # Check that the pre_page_move signal was fired
        self.assertEqual(pre_moved_handler.call_count, 1)
        pre_moved_handler.assert_called_with(
            signal=mock.ANY,
            sender=self.test_page_a.specific_class,
            instance=self.test_page_a,
            parent_page_before=nonspecific_section_a,
            parent_page_after=self.section_b,
            url_path_before="/home/section-a/hello-world/",
            url_path_after="/home/section-b/hello-world/",
        )

        # Check that the post_page_move signal was fired
        self.assertEqual(post_moved_handler.call_count, 1)
        post_moved_handler.assert_called_with(
            signal=mock.ANY,
            sender=self.test_page_a.specific_class,
            # during the move operation, we reloaded the page as a non-specific instance
            instance=Page.objects.get(pk=self.test_page_a.pk),
            parent_page_before=nonspecific_section_a,
            parent_page_after=self.section_b,
            url_path_before="/home/section-a/hello-world/",
            url_path_after="/home/section-b/hello-world/",
        )

    def test_before_move_page_hook(self):
        def hook_func(request, page, destination):
            self.assertIsInstance(request, HttpRequest)
            self.assertIsInstance(page.specific, SimplePage)
            self.assertIsInstance(destination.specific, SimplePage)

            return HttpResponse("Overridden!")

        with self.register_hook("before_move_page", hook_func):
            response = self.client.get(
                reverse(
                    "wagtailadmin_pages:move_confirm",
                    args=(self.test_page_a.id, self.section_b.id),
                )
            )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"Overridden!")

    def test_before_move_page_hook_post(self):
        def hook_func(request, page, destination):
            self.assertIsInstance(request, HttpRequest)
            self.assertIsInstance(page.specific, SimplePage)
            self.assertIsInstance(destination.specific, SimplePage)

            return HttpResponse("Overridden!")

        with self.register_hook("before_move_page", hook_func):
            response = self.client.post(
                reverse(
                    "wagtailadmin_pages:move_confirm",
                    args=(self.test_page_a.id, self.section_b.id),
                )
            )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"Overridden!")

        # page should not be moved
        self.assertEqual(
            Page.objects.get(id=self.test_page_a.id).get_parent().id, self.section_a.id
        )

    def test_after_move_page_hook(self):
        def hook_func(request, page):
            self.assertIsInstance(request, HttpRequest)
            self.assertIsInstance(page.specific, SimplePage)

            return HttpResponse("Overridden!")

        with self.register_hook("after_move_page", hook_func):
            response = self.client.post(
                reverse(
                    "wagtailadmin_pages:move_confirm",
                    args=(self.test_page_a.id, self.section_b.id),
                )
            )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"Overridden!")

        # page should be moved
        self.assertEqual(
            Page.objects.get(id=self.test_page_a.id).get_parent().id, self.section_b.id
        )

    def test_page_move_after_parent_page_types_changes_to_different_parent_model(self):
        # Test for issue #10348
        # While BusinessSubIndex cannot be created under a SimplePage, we can
        # still create it under a SimplePage invoking django-treebeard's add_child
        # which works great for our purposes.
        self.assertFalse(BusinessSubIndex.can_exist_under(self.section_a))
        page = self.section_a.add_child(
            instance=BusinessSubIndex(
                title="Business Sub Index",
                slug="business-sub-index",
                live=True,
                has_unpublished_changes=False,
            )
        )

        response = self.client.get(reverse("wagtailadmin_pages:move", args=(page.id,)))
        self.assertEqual(response.status_code, 200)

        form = response.context["move_form"]
        self.assertEqual(form.fields["new_parent_page"].initial.pk, self.section_a.pk)
        self.assertNotContains(response, self.section_a.title)



# DIRECT_FORM_FIELD_OVERRIDES, FORM_FIELD_OVERRIDES are imported for backwards
# compatibility, as people are likely importing them from here and then
# appending their own overrides
from wagtail.admin.forms.models import (  # NOQA: F401
    DIRECT_FORM_FIELD_OVERRIDES,
    FORM_FIELD_OVERRIDES,
)

from .base import *  # NOQA: F403
from .comment_panel import *  # NOQA: F403
from .field_panel import *  # NOQA: F403
from .group import *  # NOQA: F403
from .help_panel import *  # NOQA: F403
from .inline_panel import *  # NOQA: F403
from .model_utils import *  # NOQA: F403
from .multiple_chooser_panel import *  # NOQA: F403
from .page_chooser_panel import *  # NOQA: F403
from .page_utils import *  # NOQA: F403
from .publishing_panel import *  # NOQA: F403
from .signal_handlers import *  # NOQA: F403
from .title_field_panel import *  # NOQA: F403



from django.contrib.postgres.search import SearchQueryField
from django.db.models.expressions import Expression, Value


class LexemeCombinable(Expression):
    BITAND = "&"
    BITOR = "|"

    def _combine(self, other, connector, reversed, node=None):
        if not isinstance(other, LexemeCombinable):
            raise TypeError(
                f"Lexeme can only be combined with other Lexemes, got {type(other)}."
            )
        if reversed:
            return CombinedLexeme(other, connector, self)
        return CombinedLexeme(self, connector, other)

    # On Combinable, these are not implemented to reduce confusion with Q. In
    # this case we are actually (ab)using them to do logical combination so
    # it's consistent with other usage in Django.
    def bitand(self, other):
        return self._combine(other, self.BITAND, False)

    def bitor(self, other):
        return self._combine(other, self.BITOR, False)

    def __or__(self, other):
        return self._combine(other, self.BITOR, False)

    def __and__(self, other):
        return self._combine(other, self.BITAND, False)


class Lexeme(LexemeCombinable, Value):
    _output_field = SearchQueryField()

    def __init__(
        self, value, output_field=None, *, invert=False, prefix=False, weight=None
    ):
        self.prefix = prefix
        self.invert = invert
        self.weight = weight
        super().__init__(value, output_field=output_field)

    def as_sql(self, compiler, connection):
        param = "'%s'" % self.value.replace("'", "''").replace("\\", "\\\\")

        template = "%s"

        label = ""
        if self.prefix:
            label += "*"
        if self.weight:
            label += self.weight

        if label:
            param = f"{param}:{label}"
        if self.invert:
            param = f"!{param}"

        return template, [param]


class CombinedLexeme(LexemeCombinable):
    _output_field = SearchQueryField()

    def __init__(self, lhs, connector, rhs, output_field=None):
        super().__init__(output_field=output_field)
        self.connector = connector
        self.lhs = lhs
        self.rhs = rhs

    def as_sql(self, compiler, connection):
        value_params = []
        lsql, params = compiler.compile(self.lhs)
        value_params.extend(params)

        rsql, params = compiler.compile(self.rhs)
        value_params.extend(params)

        combined_sql = f"({lsql} {self.connector} {rsql})"
        combined_value = combined_sql % tuple(value_params)
        return "%s", [combined_value]
